#!/usr/bin/env python

import os
import pathlib
import shlex
import shutil
import socket
import tarfile
import time
import uuid
from functools import lru_cache
from glob import glob
from typing import List

import paramiko
import paramiko.ssh_exception
from dargs.dargs import Argument

from dpdispatcher import dlog
from dpdispatcher.base_context import BaseContext

# from dpdispatcher.submission import Machine
from dpdispatcher.utils import RetrySignal, generate_totp, get_sha256, retry, rsync


class SSHSession:
    def __init__(
        self,
        hostname,
        username,
        password=None,
        port=22,
        key_filename=None,
        passphrase=None,
        timeout=10,
        totp_secret=None,
        tar_compress=True,
        look_for_keys=True,
    ):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.key_filename = key_filename
        self.passphrase = passphrase
        self.timeout = timeout
        self.totp_secret = totp_secret
        self.ssh = None
        self.tar_compress = tar_compress
        self.look_for_keys = look_for_keys
        self._setup_ssh()

    # @classmethod
    # def deserialize(cls, jdata):
    #     instance = cls(**jdata)
    #     return instance

    # def bk_ensure_alive(self,
    #                  max_check = 10,
    #                  sleep_time = 10):
    #     count = 1
    #     while not self._check_alive():
    #         if count == max_check:
    #             raise RuntimeError('cannot connect ssh after %d failures at interval %d s' %
    #                                (max_check, sleep_time))
    #         dlog.info('connection check failed, try to reconnect to ' + self.remote_host)
    #         self._setup_ssh(hostname=self.remote_host,
    #                         port=self.remote_port,
    #                         username=self.remote_uname,
    #                         password=self.remote_password,
    #                         key_filename=self.local_key_filename,
    #                         timeout=self.remote_timeout,
    #                         passphrase=self.local_key_passphrase)
    #         count += 1
    #         time.sleep(sleep_time)

    def ensure_alive(self, max_check=10, sleep_time=10):
        count = 1
        while not self._check_alive():
            if count == max_check:
                raise RuntimeError(
                    "cannot connect ssh after %d failures at interval %d s"
                    % (max_check, sleep_time)
                )
            dlog.info("connection check failed, try to reconnect to " + self.hostname)
            self._setup_ssh()
            count += 1
            time.sleep(sleep_time)

    def _check_alive(self):
        if self.ssh is None:
            return False
        try:
            transport = self.ssh.get_transport()
            assert transport is not None
            transport.send_ignore()
            return True
        except EOFError:
            return False

    # def bk_setup_ssh(self,
    #                hostname,
    #                port=22,
    #                username=None,
    #                password=None,
    #                key_filename=None,
    #                timeout=None,
    #                passphrase=None):
    #     self.ssh = paramiko.SSHClient()
    #     # ssh_client.load_system_host_keys()
    #     self.ssh.set_missing_host_key_policy(paramiko.WarningPolicy)
    #     self.ssh.connect(hostname=hostname, port=port,
    #                      username=username, password=password,
    #                      key_filename=key_filename, timeout=timeout, passphrase=passphrase)
    #     assert(self.ssh.get_transport().is_active())
    #     transport = self.ssh.get_transport()
    #     transport.set_keepalive(60)

    @retry(max_retry=6, sleep=1)
    def _setup_ssh(self):
        # machine = self.machine
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy)
        # if self.totp_secret and self.password is None:
        #     self.password = generate_totp(self.totp_secret)
        # self.ssh.connect(hostname=self.hostname, port=self.port,
        #                 username=self.username, password=self.password,
        #                 key_filename=self.key_filename, timeout=self.timeout,passphrase=self.passphrase,
        #                 compress=True,
        #                 )
        # assert(self.ssh.get_transport().is_active())
        # transport = self.ssh.get_transport()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        sock.connect((self.hostname, self.port))

        # Make a Paramiko Transport object using the socket
        ts = paramiko.Transport(sock)
        ts.banner_timeout = 60
        ts.use_compression(compress=True)

        # Tell Paramiko that the Transport is going to be used as a client
        ts.start_client(timeout=self.timeout)

        # Begin authentication; note that the username and callback are passed
        key = None
        key_ok = False
        key_error = None
        keyfiles = []
        if self.key_filename:
            key_path = os.path.abspath(self.key_filename)
            if os.path.exists(key_path):
                for pkey_class in (
                    paramiko.RSAKey,
                    paramiko.DSSKey,
                    paramiko.ECDSAKey,
                    paramiko.Ed25519Key,
                ):
                    try:
                        # passing empty passphrase would not raise error.
                        key = pkey_class.from_private_key_file(
                            key_path, self.passphrase
                        )
                    except paramiko.SSHException as e:
                        pass
                    if key is not None:
                        break
        elif self.look_for_keys:
            for keytype, name in [
                (paramiko.RSAKey, "rsa"),
                (paramiko.DSSKey, "dsa"),
                (paramiko.ECDSAKey, "ecdsa"),
                (paramiko.Ed25519Key, "ed25519"),
            ]:
                for directory in [".ssh", "ssh"]:
                    full_path = os.path.join(
                        os.path.expanduser("~"), directory, f"id_{name}"
                    )
                    if os.path.isfile(full_path):
                        keyfiles.append((keytype, full_path))
                        # TODO: supporting cert
            for pkey_class, filename in keyfiles:
                try:
                    key = pkey_class.from_private_key_file(filename, self.passphrase)
                except paramiko.SSHException as e:
                    pass
                if key is not None:
                    break

        if key is not None:
            try:
                ts.auth_publickey(self.username, key)
            except paramiko.ssh_exception.AuthenticationException as e:
                key_error = e
            else:
                key_ok = True
        if self.totp_secret is not None:
            try:
                ts.auth_interactive(self.username, self.inter_handler)
            except paramiko.ssh_exception.AuthenticationException:
                # since the asynchrony of interactive authentication, one addtional try is added
                # retry for up to 6 times
                raise RetrySignal("Authentication failed")
        elif key_ok:
            pass
        elif self.password is not None:
            ts.auth_password(self.username, self.password)
        elif key_error is not None:
            raise RuntimeError(
                "Authentication failed, try to provide password"
            ) from key_error
        else:
            raise RuntimeError("Please provide at least one form of authentication")
        assert ts.is_active()
        # Opening a session creates a channel along the socket to the server
        try:
            ts.open_session(timeout=self.timeout)
        except paramiko.ssh_exception.SSHException:
            # retry for up to 6 times
            # ref: https://github.com/paramiko/paramiko/issues/1508
            raise RetrySignal("Opening session failed")
        ts.set_keepalive(60)
        self.ssh._transport = ts  # type: ignore
        # reset sftp
        self._sftp = None

    def inter_handler(self, title, instructions, prompt_list):
        """inter_handler: the callback for paramiko.transport.auth_interactive.

        The prototype for this function is defined by Paramiko, so all of the
        arguments need to be there, even though we don't use 'title' or
        'instructions'.

        The function is expected to return a tuple of data containing the
        responses to the provided prompts. Experimental results suggests that
        there will be one call of this function per prompt, but the mechanism
        allows for multiple prompts to be sent at once, so it's best to assume
        that that can happen.

        Since tuples can't really be built on the fly, the responses are
        collected in a list which is then converted to a tuple when it's time
        to return a value.

        Experiments suggest that the username prompt never happens. This makes
        sense, but the Username prompt is included here just in case.
        """
        resp = []  # Initialize the response container

        # Walk the list of prompts that the server sent that we need to answer
        for pr in prompt_list:
            # str() used to to make sure that we're dealing with a string rather than a unicode string
            # strip() used to get rid of any padding spaces sent by the server
            pr_str = str(pr[0]).strip().lower()
            if "username" in pr_str:
                resp.append(self.username)
            elif "password" in pr_str:
                resp.append(self.password)
            elif (
                "verification" in pr_str
                or "token" in pr_str
                and self.totp_secret is not None
            ):
                assert self.totp_secret is not None
                resp.append(generate_totp(self.totp_secret))

        return resp

    def get_ssh_client(self):
        return self.ssh

    # def get_session_root(self):
    #     return self.remote_root

    def close(self):
        assert self.ssh is not None
        self.ssh.close()

    @retry(sleep=1)
    def exec_command(self, cmd):
        """Calling self.ssh.exec_command but has an exception check."""
        assert self.ssh is not None
        try:
            return self.ssh.exec_command(cmd)
        except (paramiko.ssh_exception.SSHException, socket.timeout) as e:
            # SSH session not active
            # retry for up to 3 times
            # ensure alive
            self.ensure_alive()
            raise RetrySignal("SSH session not active in calling %s" % cmd) from e

    @property
    def sftp(self):
        """Returns sftp. Open a new one if not existing."""
        if self._sftp is None:
            assert self.ssh is not None
            self.ensure_alive()
            self._sftp = self.ssh.open_sftp()
        return self._sftp

    @staticmethod
    def arginfo():
        doc_hostname = "hostname or ip of ssh connection."
        doc_username = "username of target linux system"
        doc_password = (
            "(deprecated) password of linux system. Please use "
            "`SSH keys <https://www.ssh.com/academy/ssh/key>`_ instead to improve security."
        )
        doc_port = "ssh connection port."
        doc_key_filename = (
            "key filename used by ssh connection. If left None, find key in ~/.ssh or "
            "use password for login"
        )
        doc_passphrase = "passphrase of key used by ssh connection"
        doc_timeout = "timeout of ssh connection"
        doc_totp_secret = (
            "Time-based one time password secret. It should be a base32-encoded string"
            " extracted from the 2D code."
        )
        doc_tar_compress = "The archive will be compressed in upload and download if it is True. If not, compression will be skipped."
        doc_look_for_keys = (
            "enable searching for discoverable private key files in ~/.ssh/"
        )

        ssh_remote_profile_args = [
            Argument("hostname", str, optional=False, doc=doc_hostname),
            Argument("username", str, optional=False, doc=doc_username),
            Argument("password", str, optional=True, doc=doc_password),
            Argument("port", int, optional=True, default=22, doc=doc_port),
            Argument(
                "key_filename",
                [str, type(None)],
                optional=True,
                default=None,
                doc=doc_key_filename,
            ),
            Argument(
                "passphrase",
                [str, type(None)],
                optional=True,
                default=None,
                doc=doc_passphrase,
            ),
            Argument("timeout", int, optional=True, default=10, doc=doc_timeout),
            Argument(
                "totp_secret", str, optional=True, default=None, doc=doc_totp_secret
            ),
            Argument(
                "tar_compress", bool, optional=True, default=True, doc=doc_tar_compress
            ),
            Argument(
                "look_for_keys",
                bool,
                optional=True,
                default=True,
                doc=doc_look_for_keys,
            ),
        ]
        ssh_remote_profile_format = Argument(
            "ssh_session", dict, ssh_remote_profile_args
        )
        return ssh_remote_profile_format

    def put(self, from_f, to_f):
        if self.rsync_available:
            return rsync(
                from_f,
                self.remote + ":" + to_f,
                port=self.port,
                key_filename=self.key_filename,
                timeout=self.timeout,
            )
        return self.sftp.put(from_f, to_f)

    def get(self, from_f, to_f):
        if self.rsync_available:
            return rsync(
                self.remote + ":" + from_f,
                to_f,
                port=self.port,
                key_filename=self.key_filename,
                timeout=self.timeout,
            )
        return self.sftp.get(from_f, to_f)

    @property
    @lru_cache(maxsize=None)
    def rsync_available(self) -> bool:
        return (
            shutil.which("rsync") is not None
            and self.password is None
            and self.exec_command("rsync --version")[1].channel.recv_exit_status() == 0
            and self.totp_secret is None
            and self.passphrase is None
        )

    @property
    def remote(self) -> str:
        return f"{self.username}@{self.hostname}"


class SSHContext(BaseContext):
    def __init__(
        self,
        local_root,
        remote_root,
        remote_profile,
        clean_asynchronously=False,
        *args,
        **kwargs,
    ):
        assert type(local_root) == str
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        assert os.path.isabs(remote_root), "remote_root must be a abspath"
        self.temp_remote_root = remote_root
        self.remote_profile = remote_profile
        self.remote_root = None

        # self.job_uuid = None
        self.clean_asynchronously = clean_asynchronously
        # self.job_uuid = job_uuid
        # if job_uuid:
        #    self.job_uuid=job_uuid
        # else:
        #    self.job_uuid = str(uuid.uuid4())
        self.ssh_session = SSHSession(**remote_profile)  # type: ignore
        # self.temp_remote_root = os.path.join(self.ssh_session.get_session_root())
        self.ssh_session.ensure_alive()
        try:
            self.sftp.mkdir(self.temp_remote_root)
        except OSError:
            pass

    @classmethod
    def load_from_dict(cls, context_dict):
        # instance = cls()
        # input = dict(
        #     hostname = jdata['hostname'],
        #     remote_root = jdata['remote_root'],
        #     username = jdata['username'],
        #     password = jdata.get('password', None),
        #     port = jdata.get('port', 22),
        #     key_filename = jdata.get('key_filename', None),
        #     passphrase = jdata.get('passphrase', None),
        #     timeout = jdata.get('timeout', 10),
        #     tar_compress = jdata.get('tar_compress', True)
        # )
        local_root = context_dict["local_root"]
        remote_root = context_dict["remote_root"]
        remote_profile = context_dict["remote_profile"]
        clean_asynchronously = context_dict.get("clean_asynchronously", False)

        ssh_context = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
            clean_asynchronously=clean_asynchronously,
        )
        # local_root = jdata['local_root']
        # ssh_session = SSHSession(**input)
        # ssh_context = SSHContext(
        #     local_root=local_root,
        #     ssh_session=ssh_session,
        #     clean_asynchronously=jdata.get('clean_asynchronously', False),
        #     )
        return ssh_context

    @property
    def ssh(self):
        return self.ssh_session.get_ssh_client()

    @property
    def sftp(self):
        return self.ssh_session.sftp

    def close(self):
        self.ssh_session.close()

    def get_job_root(self):
        return self.remote_root

    def bind_submission(self, submission):
        assert self.ssh_session is not None
        assert self.ssh_session.ssh is not None
        self.submission = submission
        self.local_root = pathlib.PurePath(
            os.path.join(self.temp_local_root, submission.work_base)
        ).as_posix()
        old_remote_root = self.remote_root
        # self.remote_root = os.path.join(self.temp_remote_root, self.submission.submission_hash, self.submission.work_base )
        self.remote_root = pathlib.PurePath(
            os.path.join(self.temp_remote_root, self.submission.submission_hash)
        ).as_posix()
        # move the working directory if remote_root changes
        if (
            old_remote_root is not None
            and old_remote_root != self.remote_root
            and self.check_file_exists(old_remote_root)
            and not self.check_file_exists(self.remote_root)
        ):
            self.block_checkcall(
                f"mv {shlex.quote(old_remote_root)} {shlex.quote(self.remote_root)}"
            )
        elif (
            old_remote_root is not None
            and old_remote_root != self.remote_root
            and self.check_file_exists(old_remote_root)
            and not len(self.ssh_session.sftp.listdir(old_remote_root))
        ):
            # if the new directory exists and the old directory does not contain files, then move the old directory
            self._rmtree(old_remote_root)

        sftp = self.ssh_session.ssh.open_sftp()
        try:
            sftp.mkdir(self.remote_root)
        except OSError:
            pass

        # self.job_uuid = submission.submission_hash
        # dlog.debug("debug:SSHContext.bind_submission"
        #     "{submission.submission_hash}; {self.local_root}; {self.remote_root")

        # try:
        # print('self.remote_root', self.remote_root)
        # sftp = self.ssh_session.ssh.open_sftp()
        # sftp.mkdir(self.remote_root)
        # sftp.close()
        # except Exception:
        #     pass

    def _walk_directory(self, files, work_path, file_list, directory_list):
        """Convert input path to list of files and directories."""
        for jj in files:
            file_name = os.path.join(work_path, jj)
            if os.path.isfile(file_name):
                file_list.append(file_name)
            elif os.path.isdir(file_name):
                for root, dirs, files in os.walk(
                    file_name, topdown=False, followlinks=True
                ):
                    if not files:
                        directory_list.append(root)
                    for name in files:
                        file_list.append(os.path.join(root, name))
            elif glob(file_name):
                # If the file name contains a wildcard, os.path functions will fail to identify it. Use glob to get the complete list of filenames which match the wildcard.
                abs_file_list = glob(file_name)
                rel_file_list = [
                    os.path.relpath(ii, start=work_path) for ii in abs_file_list
                ]
                self._walk_directory(
                    rel_file_list, work_path, file_list, directory_list
                )
            else:
                raise RuntimeError(f"cannot find upload file {work_path} {jj}")

    def upload(
        self,
        # job_dirs,
        submission,
        # local_up_files,
        dereference=True,
    ):
        assert self.remote_root is not None
        dlog.info(f"remote path: {self.remote_root}")
        # remote_cwd =
        self.ssh_session.sftp.chdir(self.temp_remote_root)
        recover = False
        try:
            self.ssh_session.sftp.mkdir(os.path.basename(self.remote_root))
        except OSError:
            # mkdir failed meaning it exists
            if len(self.ssh_session.sftp.listdir(os.path.basename(self.remote_root))):
                recover = True
        self.ssh_session.sftp.chdir(None)

        file_list = []
        directory_list = []
        for task in submission.belonging_tasks:
            directory_list.append(os.path.join(self.local_root, task.task_work_path))
            #     file_list.append(ii)
            self._walk_directory(
                task.forward_files,
                os.path.join(self.local_root, task.task_work_path),
                file_list,
                directory_list,
            )
        self._walk_directory(
            submission.forward_common_files, self.local_root, file_list, directory_list
        )

        # convert to relative path to local_root
        directory_list = [os.path.relpath(jj, self.local_root) for jj in directory_list]

        # check if the same file exists on the remote file
        # only check sha256 when the job is recovered
        if recover:
            # generate local sha256 file
            sha256_list = []
            for jj in file_list:
                sha256 = get_sha256(jj)
                jj_rel = pathlib.PurePath(
                    os.path.relpath(jj, self.local_root)
                ).as_posix()
                sha256_list.append(f"{sha256}  {jj_rel}")
            # write to remote
            sha256_file = os.path.join(
                self.remote_root, ".tmp.sha256." + str(uuid.uuid4())
            )
            self.write_file(sha256_file, "\n".join(sha256_list))
            # check sha256
            # `:` means pass: https://stackoverflow.com/a/2421592/9567349
            _, stdout, _ = self.block_checkcall(
                "sha256sum -c %s --quiet >.sha256sum_stdout 2>/dev/null || :"
                % shlex.quote(sha256_file)
            )
            self.sftp.remove(sha256_file)
            # regenerate file list
            file_list = []

            for ii in self.read_file(".sha256sum_stdout").split("\n"):
                if ii:
                    file_list.append(ii.split(":")[0])
        else:
            # convert to relative path to local_root
            file_list = [os.path.relpath(jj, self.local_root) for jj in file_list]

        self._put_files(
            file_list,
            dereference=dereference,
            directories=directory_list,
            tar_compress=self.remote_profile.get("tar_compress", None),
        )

    def download(
        self,
        submission,
        # job_dirs,
        # remote_down_files,
        check_exists=False,
        mark_failure=True,
        back_error=False,
    ):
        self.ssh_session.ensure_alive()
        file_list = []
        # for ii in job_dirs :
        for task in submission.belonging_tasks:
            for jj in task.backward_files:
                file_name = pathlib.PurePath(
                    os.path.join(task.task_work_path, jj)
                ).as_posix()
                if check_exists:
                    if self.check_file_exists(file_name):
                        file_list.append(file_name)
                    elif mark_failure:
                        with open(
                            os.path.join(
                                self.local_root,
                                task.task_work_path,
                                "tag_failure_download_%s" % jj,
                            ),
                            "w",
                        ) as fp:
                            pass
                    else:
                        pass
                else:
                    file_list.append(file_name)
            if back_error:
                errors = glob(os.path.join(task.task_work_path, "error*"))
                file_list.extend(errors)
        file_list.extend(submission.backward_common_files)
        if len(file_list) > 0:
            self._get_files(
                file_list, tar_compress=self.remote_profile.get("tar_compress", None)
            )

    def block_checkcall(self, cmd, asynchronously=False, stderr_whitelist=None):
        """Run command with arguments. Wait for command to complete. If the return code
        was zero then return, otherwise raise RuntimeError.

        Parameters
        ----------
        cmd : str
            The command to run.
        asynchronously : bool, optional, default=False
            Run command asynchronously. If True, `nohup` will be used to run the command.
        stderr_whitelist : list of str, optional, default=None
            If not None, the stderr will be checked against the whitelist. If the stderr
            contains any of the strings in the whitelist, the command will be considered
            successful.
        """
        assert self.remote_root is not None
        self.ssh_session.ensure_alive()
        if asynchronously:
            cmd = "nohup %s >/dev/null &" % cmd
        stdin, stdout, stderr = self.ssh_session.exec_command(
            ("cd %s ;" % shlex.quote(self.remote_root)) + cmd
        )
        exit_status = stdout.channel.recv_exit_status()
        if exit_status != 0:
            raise RuntimeError(
                "Get error code %d in calling %s through ssh with job: %s . message: %s"
                % (
                    exit_status,
                    cmd,
                    self.submission.submission_hash,
                    stderr.read().decode("utf-8"),
                )
            )
        return stdin, stdout, stderr

    def block_call(self, cmd):
        assert self.remote_root is not None
        self.ssh_session.ensure_alive()
        stdin, stdout, stderr = self.ssh_session.exec_command(
            ("cd %s ;" % shlex.quote(self.remote_root)) + cmd
        )
        exit_status = stdout.channel.recv_exit_status()
        return exit_status, stdin, stdout, stderr

    def clean(self):
        self.ssh_session.ensure_alive()
        self._rmtree(self.remote_root)

    def write_file(self, fname, write_str):
        assert self.remote_root is not None
        self.ssh_session.ensure_alive()
        fname = pathlib.PurePath(os.path.join(self.remote_root, fname)).as_posix()
        # to prevent old file from being overwritten but cancelled, create a temporary file first
        # when it is fully written, rename it to the original file name
        with self.sftp.open(fname + "~", "w") as fp:
            fp.write(write_str)
        # sftp.rename may throw OSError
        self.block_checkcall(
            "mv {} {}".format(shlex.quote(fname + "~"), shlex.quote(fname))
        )

    def read_file(self, fname):
        assert self.remote_root is not None
        self.ssh_session.ensure_alive()
        with self.sftp.open(
            pathlib.PurePath(os.path.join(self.remote_root, fname)).as_posix(), "r"
        ) as fp:
            ret = fp.read().decode("utf-8")
        return ret

    def check_file_exists(self, fname):
        assert self.remote_root is not None
        self.ssh_session.ensure_alive()
        try:
            self.sftp.stat(
                pathlib.PurePath(os.path.join(self.remote_root, fname)).as_posix()
            )
            ret = True
        except OSError:
            ret = False
        return ret

    def call(self, cmd):
        stdin, stdout, stderr = self.ssh_session.exec_command(cmd)
        # stdin, stdout, stderr = self.ssh.exec_command('echo $$; exec ' + cmd)
        # pid = stdout.readline().strip()
        # print(pid)
        return {"stdin": stdin, "stdout": stdout, "stderr": stderr}

    def check_finish(self, cmd_pipes):
        return cmd_pipes["stdout"].channel.exit_status_ready()

    def get_return(self, cmd_pipes):
        if not self.check_finish(cmd_pipes):
            return None, None, None
        else:
            retcode = cmd_pipes["stdout"].channel.recv_exit_status()
            return retcode, cmd_pipes["stdout"], cmd_pipes["stderr"]

    def _rmtree(self, remotepath, verbose=False):
        """Remove the remote path."""
        # The original implementation method removes files one by one using sftp.
        # If the latency of the remote server is high, it is very slow.
        # Thus, it's better to use system's `rm` to remove a directory, which may
        # save a lot of time.
        if verbose:
            dlog.info("removing %s" % remotepath)
        # In some supercomputers, it's very slow to remove large numbers of files
        # (e.g. directory containing trajectory) due to bad I/O performance.
        # So an asynchronously option is provided.
        self.block_checkcall(
            "rm -rf %s" % shlex.quote(remotepath),
            asynchronously=self.clean_asynchronously,
        )

    def _put_files(
        self,
        files,
        dereference=True,
        directories=None,
        tar_compress=True,
    ):
        """Upload files to server.

        Parameters
        ----------
        files : list
            uploaded files
        dereference : bool, default: True
            If dereference is False, add symbolic and hard links to the archive.
            If it is True, add the content of the target files to the archive.
            This has no effect on systems that do not support symbolic links.
        directories : list, default: None
            uploaded directories non-recursively. Use `files` for uploading
            recursively
        tar_compress : bool, default: True
            If tar_compress is True, compress the archive using gzip
            It it is False, then it is uncompressed
        """
        assert self.remote_root is not None
        of_suffix = ".tgz"
        tarfile_mode = "w:gz"
        kwargs = {"compresslevel": 6}
        if not tar_compress:
            of_suffix = ".tar"
            tarfile_mode = "w"
            kwargs = {}

        of = self.submission.submission_hash + of_suffix
        # local tar
        if os.path.isfile(os.path.join(self.local_root, of)):
            os.remove(os.path.join(self.local_root, of))
        with tarfile.open(
            os.path.join(self.local_root, of),
            tarfile_mode,
            dereference=dereference,
            **kwargs,
        ) as tar:
            # avoid compressing duplicated files or directories
            for ii in set(files):
                ii_full = os.path.join(self.local_root, ii)
                tar.add(ii_full, arcname=ii)
            if directories is not None:
                for ii in set(directories):
                    ii_full = os.path.join(self.local_root, ii)
                    tar.add(ii_full, arcname=ii, recursive=False)
        self.ssh_session.ensure_alive()
        try:
            self.sftp.mkdir(self.remote_root)
        except OSError:
            pass
        # trans
        from_f = pathlib.PurePath(os.path.join(self.local_root, of)).as_posix()
        to_f = pathlib.PurePath(os.path.join(self.remote_root, of)).as_posix()
        try:
            self.ssh_session.put(from_f, to_f)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"from {from_f} to {self.ssh_session.username} @ {self.ssh_session.hostname} : {to_f} Error!"
            )
        # remote extract
        self.block_checkcall("tar xf %s" % of)
        # clean up
        os.remove(from_f)
        self.sftp.remove(to_f)

    def _get_files(self, files, tar_compress=True):
        assert self.remote_root is not None
        # avoid compressing duplicated files
        files = list(set(files))

        of_suffix = ".tar.gz"
        tarfile_mode = "r:gz"
        tar_command = "czfh"
        if not tar_compress:
            of_suffix = ".tar"
            tarfile_mode = "r"
            tar_command = "cfh"

        of = self.submission.submission_hash + of_suffix
        # remote tar
        # If the number of files are large, we may get "Argument list too long" error.
        # Thus, "-T" accepts a file containing the list of files
        per_nfile = 100
        ntar = len(files) // per_nfile + 1
        if ntar <= 1:
            self.block_checkcall(
                "tar {} {} {}".format(
                    tar_command,
                    shlex.quote(of),
                    " ".join([shlex.quote(file) for file in files]),
                )
            )
        else:
            file_list_file = os.path.join(
                self.remote_root, ".tmp.tar." + str(uuid.uuid4())
            )
            self.write_file(file_list_file, "\n".join(files))
            self.block_checkcall(
                f"tar {tar_command} {shlex.quote(of)} -T {shlex.quote(file_list_file)}"
            )
        # trans
        from_f = pathlib.PurePath(os.path.join(self.remote_root, of)).as_posix()
        to_f = pathlib.PurePath(os.path.join(self.local_root, of)).as_posix()
        if os.path.isfile(to_f):
            os.remove(to_f)
        self.ssh_session.get(from_f, to_f)
        # extract
        with tarfile.open(to_f, mode=tarfile_mode) as tar:
            tar.extractall(path=self.local_root)
        # cleanup
        os.remove(to_f)
        self.sftp.remove(from_f)

    @classmethod
    def machine_subfields(cls) -> List[Argument]:
        """Generate the machine subfields.

        Returns
        -------
        list[Argument]
            machine subfields
        """
        doc_remote_profile = (
            "The information used to maintain the connection with remote machine."
        )
        remote_profile_format = SSHSession.arginfo()
        remote_profile_format.name = "remote_profile"
        remote_profile_format.doc = doc_remote_profile
        return [remote_profile_format]



import hashlib
import os
import shutil
import subprocess as sp
from glob import glob
from subprocess import TimeoutExpired

from dpdispatcher import dlog
from dpdispatcher.base_context import BaseContext


class SPRetObj:
    def __init__(self, ret):
        self.data = ret

    def read(self):
        return self.data

    def readlines(self):
        lines = self.data.decode("utf-8").splitlines()
        ret = []
        for aa in lines:
            ret.append(aa + "\n")
        return ret


def _check_file_path(fname):
    dirname = os.path.dirname(fname)
    if dirname != "":
        os.makedirs(dirname, exist_ok=True)


def _identical_files(fname0, fname1):
    with open(fname0) as fp:
        code0 = hashlib.sha1(fp.read().encode("utf-8")).hexdigest()
    with open(fname1) as fp:
        code1 = hashlib.sha1(fp.read().encode("utf-8")).hexdigest()
    return code0 == code1


class LocalContext(BaseContext):
    """Run jobs in the local server and remote directory.

    Parameters
    ----------
    local_root : str
        The local directory to store the jobs.
    remote_root : str
        The remote directory to store the jobs.
    remote_profile : dict, optional
        The remote profile. The default is {}.
    *args
        The arguments.
    **kwargs
        The keyword arguments.
    """

    def __init__(
        self,
        local_root,
        remote_root,
        remote_profile={},
        *args,
        **kwargs,
    ):
        assert type(local_root) == str
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.temp_remote_root = os.path.abspath(remote_root)
        self.remote_profile = remote_profile

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict["local_root"]
        remote_root = context_dict["remote_root"]
        remote_profile = context_dict.get("remote_profile", {})
        instance = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
        )
        return instance

    def get_job_root(self):
        return self.remote_root

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = os.path.join(
            self.temp_remote_root, submission.submission_hash
        )

    def upload(self, submission):
        os.makedirs(self.remote_root, exist_ok=True)
        for ii in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, ii.task_work_path)
            remote_job = os.path.join(self.remote_root, ii.task_work_path)
            os.makedirs(remote_job, exist_ok=True)

            file_list = []
            for kk in ii.forward_files:
                abs_file_list = glob(os.path.join(local_job, kk))
                if not abs_file_list:
                    raise RuntimeError(
                        "cannot find upload file " + os.path.join(local_job, kk)
                    )
                rel_file_list = [
                    os.path.relpath(ii, start=local_job) for ii in abs_file_list
                ]
                file_list.extend(rel_file_list)

            for jj in file_list:
                if not os.path.exists(os.path.join(local_job, jj)):
                    raise RuntimeError(
                        "cannot find upload file " + os.path.join(local_job, jj)
                    )
                if os.path.exists(os.path.join(remote_job, jj)):
                    os.remove(os.path.join(remote_job, jj))
                _check_file_path(os.path.join(remote_job, jj))
                os.symlink(os.path.join(local_job, jj), os.path.join(remote_job, jj))

        local_job = self.local_root
        remote_job = self.remote_root

        file_list = []
        for kk in submission.forward_common_files:
            abs_file_list = glob(os.path.join(local_job, kk))
            if not abs_file_list:
                raise RuntimeError(
                    "cannot find upload file " + os.path.join(local_job, kk)
                )
            rel_file_list = [
                os.path.relpath(ii, start=local_job) for ii in abs_file_list
            ]
            file_list.extend(rel_file_list)

        for jj in file_list:
            if not os.path.exists(os.path.join(local_job, jj)):
                raise RuntimeError(
                    "cannot find upload file " + os.path.join(local_job, jj)
                )
            if os.path.exists(os.path.join(remote_job, jj)):
                os.remove(os.path.join(remote_job, jj))
            _check_file_path(os.path.join(remote_job, jj))
            os.symlink(os.path.join(local_job, jj), os.path.join(remote_job, jj))

    def download(
        self, submission, check_exists=False, mark_failure=True, back_error=False
    ):
        for ii in submission.belonging_tasks:
            local_job = os.path.join(self.local_root, ii.task_work_path)
            remote_job = os.path.join(self.remote_root, ii.task_work_path)
            flist = ii.backward_files
            if back_error:
                flist += glob(os.path.join(remote_job, "error*"))
            for jj in flist:
                rfile = os.path.join(remote_job, jj)
                lfile = os.path.join(local_job, jj)
                if not os.path.realpath(rfile) == os.path.realpath(lfile):
                    if (not os.path.exists(rfile)) and (not os.path.exists(lfile)):
                        if check_exists:
                            if mark_failure:
                                tag_file_path = os.path.join(
                                    self.local_root,
                                    ii.task_work_path,
                                    "tag_failure_download_%s" % jj,
                                )
                                with open(tag_file_path, "w") as fp:
                                    pass
                            else:
                                pass
                        else:
                            raise RuntimeError("do not find download file " + rfile)
                    elif (not os.path.exists(rfile)) and (os.path.exists(lfile)):
                        # already downloaded
                        pass
                    elif (os.path.exists(rfile)) and (not os.path.exists(lfile)):
                        # trivial case, download happily
                        # for links, copy instead of moving (default behavior of copyfile is following symlinks)
                        if not os.path.islink(rfile):
                            shutil.move(rfile, lfile)
                        else:
                            shutil.copyfile(rfile, lfile)
                    elif (os.path.exists(rfile)) and (os.path.exists(lfile)):
                        # both exists, replace!
                        dlog.info(f"find existing {lfile}, replacing by {rfile}")
                        if os.path.isdir(lfile):
                            shutil.rmtree(lfile, ignore_errors=True)
                        elif os.path.isfile(lfile) or os.path.islink(lfile):
                            os.remove(lfile)
                        shutil.copyfile(rfile, lfile)
                    else:
                        raise RuntimeError("should not reach here!")
                else:
                    # no nothing in the case of linked files
                    pass
        local_job = self.local_root
        remote_job = self.remote_root
        flist = submission.backward_common_files
        if back_error:
            flist += glob(os.path.join(remote_job, "error*"))
        for jj in flist:
            rfile = os.path.join(remote_job, jj)
            lfile = os.path.join(local_job, jj)
            if not os.path.realpath(rfile) == os.path.realpath(lfile):
                if (not os.path.exists(rfile)) and (not os.path.exists(lfile)):
                    if check_exists:
                        if mark_failure:
                            with open(
                                os.path.join(
                                    self.local_root, "tag_failure_download_%s" % jj
                                ),
                                "w",
                            ) as fp:
                                pass
                        else:
                            pass
                    else:
                        raise RuntimeError("do not find download file " + rfile)
                elif (not os.path.exists(rfile)) and (os.path.exists(lfile)):
                    # already downloaded
                    pass
                elif (os.path.exists(rfile)) and (not os.path.exists(lfile)):
                    # trivial case, download happily
                    if not os.path.islink(rfile):
                        shutil.move(rfile, lfile)
                    else:
                        shutil.copyfile(rfile, lfile)
                elif (os.path.exists(rfile)) and (os.path.exists(lfile)):
                    dlog.info(f"both exist rfile:{rfile}; lfile:{lfile}")
                    # both exists, replace!
                    dlog.info(f"find existing {lfile}, replacing by {rfile}")
                    if os.path.isdir(lfile):
                        shutil.rmtree(lfile, ignore_errors=True)
                    elif os.path.isfile(lfile) or os.path.islink(lfile):
                        os.remove(lfile)
                    shutil.copyfile(rfile, lfile)
                else:
                    raise RuntimeError("should not reach here!")
            else:
                # no nothing in the case of linked files
                pass

    def block_checkcall(self, cmd):
        proc = sp.Popen(
            cmd, cwd=self.remote_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        o, e = proc.communicate()
        stdout = SPRetObj(o)
        stderr = SPRetObj(e)
        code = proc.returncode
        if code != 0:
            raise RuntimeError(
                "Get error code %d in locally calling %s with job: %s ",
                (code, cmd, self.submission.submission_hash),
            )
        return None, stdout, stderr

    def block_call(self, cmd):
        proc = sp.Popen(
            cmd, cwd=self.remote_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        o, e = proc.communicate()
        stdout = SPRetObj(o)
        stderr = SPRetObj(e)
        code = proc.returncode
        return code, None, stdout, stderr

    def clean(self):
        shutil.rmtree(self.remote_root, ignore_errors=True)

    def write_file(self, fname, write_str):
        os.makedirs(self.remote_root, exist_ok=True)
        with open(os.path.join(self.remote_root, fname), "w") as fp:
            fp.write(write_str)

    def read_file(self, fname):
        with open(os.path.join(self.remote_root, fname)) as fp:
            ret = fp.read()
        return ret

    def check_file_exists(self, fname):
        return os.path.isfile(os.path.join(self.remote_root, fname))

    def call(self, cmd):
        proc = sp.Popen(
            cmd, cwd=self.remote_root, shell=True, stdout=sp.PIPE, stderr=sp.PIPE
        )
        return proc

    def check_finish(self, proc):
        return proc.poll() is not None

    def get_return(self, proc):
        ret = proc.poll()
        if ret is None:
            return None, None, None
        else:
            try:
                o, e = proc.communicate()
                stdout = SPRetObj(o)
                stderr = SPRetObj(e)
            except TimeoutExpired:
                stdout = None
                stderr = None
        return ret, stdout, stderr



import hashlib
import os
import pathlib
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

import dpdispatcher  # noqa: F401
from dpdispatcher.base_context import BaseContext  # noqa: F401
from dpdispatcher.distributed_shell import DistributedShell  # noqa: F401
from dpdispatcher.dp_cloud_server import Lebesgue  # noqa: F401
from dpdispatcher.hdfs_cli import HDFS  # noqa: F401
from dpdispatcher.hdfs_context import HDFSContext  # noqa: F401

# from dpgen.dispatcher.Dispatcher import FinRecord
from dpdispatcher.JobStatus import JobStatus  # noqa: F401

# from dpdispatcher.local_context import local_context
from dpdispatcher.lazy_local_context import LazyLocalContext  # noqa: F401

# from dpdispatcher.local_context import LocalSession
from dpdispatcher.local_context import LocalContext, _identical_files  # noqa: F401
from dpdispatcher.lsf import LSF  # noqa: F401
from dpdispatcher.pbs import PBS  # noqa: F401
from dpdispatcher.shell import Shell  # noqa: F401
from dpdispatcher.slurm import Slurm  # noqa: F401
from dpdispatcher.ssh_context import SSHContext, SSHSession  # noqa: F401

try:
    from dpdispatcher.dp_cloud_server import DpCloudServer  # noqa: F401
    from dpdispatcher.dp_cloud_server_context import DpCloudServerContext  # noqa: F401
except Exception:
    pass
import dargs  # noqa: F401

from dpdispatcher.machine import Machine  # noqa: F401
from dpdispatcher.submission import Job, Resources, Submission, Task  # noqa: F401
from dpdispatcher.utils import RetrySignal, retry  # noqa: F401


def setUpModule():
    os.chdir(os.path.abspath(os.path.dirname(__file__)))


def get_file_md5(file_path):
    return hashlib.md5(pathlib.Path(file_path).read_bytes()).hexdigest()



import os
import shutil
import sys
import time
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "shell"
from .context import (
    JobStatus,
    LocalContext,
    LocalSession,
    Shell,
    my_file_cmp,
    setUpModule,  # noqa: F401
)


class TestShell(unittest.TestCase):
    def setUp(self):
        os.makedirs("loc", exist_ok=True)
        os.makedirs("rmt", exist_ok=True)
        os.makedirs("loc/task0", exist_ok=True)
        os.makedirs("loc/task1", exist_ok=True)
        for ii in ["loc/task0", "loc/task1"]:
            with open(os.path.join(ii, "test0"), "w") as fp:
                fp.write(str(uuid.uuid4()))
        work_profile = LocalSession({"work_path": "rmt"})
        self.ctx = LocalContext("loc", work_profile)
        self.shell = Shell(self.ctx)

    def tearDown(self):
        shutil.rmtree("loc")
        shutil.rmtree("rmt")
        if os.path.exists("dpgen.log"):
            os.remove("dpgen.log")
        if os.path.exists("run.sub"):
            os.remove("run.sub")
        if os.path.exists("run.sub.1"):
            os.remove("run.sub.1")

    def test_manual_cuda_devices(self):
        job_dirs = ["task0", "task1"]
        res = {"manual_cuda_devices": 3}
        ret = self.shell.sub_script(job_dirs, ["touch test1", "touch test2"], res=res)
        with open("run.sub.gpu", "w") as fp:
            fp.write(ret)

    def test_manual_cuda_multiplicity(self):
        job_dirs = ["task0", "task1", "task2", "task3"]
        res = {"manual_cuda_devices": 2, "manual_cuda_multiplicity": 2}
        ret = self.shell.sub_script(job_dirs, ["touch test1", "touch test2"], res=res)
        with open("run.sub.gpu.multi", "w") as fp:
            fp.write(ret)

    def test_gen_sub_script(self):
        job_dirs = ["task0", "task1"]
        self.shell.context.upload(job_dirs, ["test0"])
        ret = self.shell.sub_script(job_dirs, ["touch test1", "touch test2"])
        with open("run.sub", "w") as fp:
            fp.write(ret)
        ret1 = self.shell.sub_script(
            job_dirs,
            ["touch", "touch"],
            args=[["test1 ", "test1 "], ["test2 ", "test2 "]],
        )
        with open("run.sub.1", "w") as fp:
            fp.write(ret1)
        time.sleep(1)
        my_file_cmp(self, "run.sub.1", "run.sub")
        # with open('run.sub', 'w') as fp:
        #     fp.write(ret)

    def test_sub_success(self):
        job_dirs = ["task0", "task1"]
        self.shell.context.upload(job_dirs, ["test0"])
        self.shell.submit(job_dirs, ["touch test1", "touch test2"])
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished:
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.shell.context.remote_root,
                    "%s_tag_finished" % self.shell.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test2")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test2")
            )
        )

    def test_sub_scancel(self):
        job_dirs = ["task0", "task1"]
        self.shell.context.upload(job_dirs, ["test0"])
        # sub
        self.shell.submit(job_dirs, ["touch test1", "sleep 10"])
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished:
                raise RuntimeError("should not finished")
            if ret == JobStatus.running:
                # wait for file writing
                time.sleep(2)
                # kill job
                self.shell.context.kill(self.shell.proc)
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.shell.context.remote_root,
                    "%s_tag_finished" % self.shell.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test1")
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test2")
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test2")
            )
        )
        # sub restart
        self.shell.submit(job_dirs, ["rm test1", "touch test2"], restart=True)
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished:
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.shell.context.remote_root,
                    "%s_tag_finished" % self.shell.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test2")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test2")
            )
        )



from dpdispatcher import dlog
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.machine import Machine
from dpdispatcher.utils import run_cmd_with_all_output

shell_script_header_template = """
#!/bin/bash -l
set -x
"""

script_env_template = """
{module_unload_part}
{module_load_part}
{source_files_part}
{export_envs_part}
{prepend_script_part}

REMOTE_ROOT=`pwd`
echo 0 > {flag_if_job_task_fail}
test $? -ne 0 && exit 1

if ! ls {submission_hash}_upload.tgz 1>/dev/null 2>&1; then
    hadoop fs -get {remote_root}/*.tgz .
fi
for TGZ in `ls *.tgz`; do tar xvf $TGZ; done

"""
script_end_template = """
cd $REMOTE_ROOT
test $? -ne 0 && exit 1

wait
FLAG_IF_JOB_TASK_FAIL=$(cat {flag_if_job_task_fail})
if test $FLAG_IF_JOB_TASK_FAIL -eq 0; then
    tar czf {submission_hash}_{job_hash}_download.tar.gz {all_task_dirs}
    hadoop fs -put -f {submission_hash}_{job_hash}_download.tar.gz {remote_root}
    hadoop fs -touchz {remote_root}/{job_tag_finished}
else
    exit 1
fi
{append_script_part}
"""


class DistributedShell(Machine):
    def gen_script_env(self, job):
        source_files_part = ""

        module_unload_part = ""
        module_purge = job.resources.module_purge
        if module_purge:
            module_unload_part += "module purge\n"
        module_unload_list = job.resources.module_unload_list
        for ii in module_unload_list:
            module_unload_part += f"module unload {ii}\n"

        module_load_part = ""
        module_list = job.resources.module_list
        for ii in module_list:
            module_load_part += f"module load {ii}\n"

        source_list = job.resources.source_list
        for ii in source_list:
            line = "{ source %s; } \n" % ii
            source_files_part += line

        export_envs_part = ""
        envs = job.resources.envs
        for k, v in envs.items():
            if isinstance(v, list):
                for each_value in v:
                    export_envs_part += f"export {k}={each_value}\n"
            else:
                export_envs_part += f"export {k}={v}\n"

        prepend_script = job.resources.prepend_script
        prepend_script_part = "\n".join(prepend_script)

        flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"

        script_env = script_env_template.format(
            flag_if_job_task_fail=flag_if_job_task_fail,
            module_unload_part=module_unload_part,
            module_load_part=module_load_part,
            source_files_part=source_files_part,
            export_envs_part=export_envs_part,
            prepend_script_part=prepend_script_part,
            remote_root=self.context.remote_root,
            submission_hash=self.context.submission.submission_hash,
        )
        return script_env

    def gen_script_end(self, job):
        all_task_dirs = ""
        for task in job.job_task_list:
            all_task_dirs += "%s " % task.task_work_path
        job_tag_finished = job.job_hash + "_job_tag_finished"
        flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"

        append_script = job.resources.append_script
        append_script_part = "\n".join(append_script)

        script_end = script_end_template.format(
            job_tag_finished=job_tag_finished,
            flag_if_job_task_fail=flag_if_job_task_fail,
            all_task_dirs=all_task_dirs,
            append_script_part=append_script_part,
            remote_root=self.context.remote_root,
            submission_hash=self.context.submission.submission_hash,
            job_hash=job.job_hash,
        )
        return script_end

    def gen_script_header(self, job):
        shell_script_header = shell_script_header_template
        return shell_script_header

    def do_submit(self, job):
        """Submit th job to yarn using distributed shell.

        Parameters
        ----------
        job : Job class instance
            job to be submitted

        Returns
        -------
        job_id: string
            submit process id
        """
        script_str = self.gen_script(job)
        script_file_name = job.script_file_name
        job_id_name = job.job_hash + "_job_id"
        output_name = job.job_hash + ".out"
        self.context.write_file(fname=script_file_name, write_str=script_str)

        resources = job.resources
        submit_command = (
            "hadoop jar {}/hadoop-yarn-applications-distributedshell-*.jar "
            "org.apache.hadoop.yarn.applications.distributedshell.Client "
            "-jar {}/hadoop-yarn-applications-distributedshell-*.jar "
            '-queue {} -appname "distributedshell_dpgen_{}" '
            "-shell_env YARN_CONTAINER_RUNTIME_TYPE=docker "
            "-shell_env YARN_CONTAINER_RUNTIME_DOCKER_IMAGE={} "
            "-shell_env ENV_DOCKER_CONTAINER_SHM_SIZE='600m' "
            "-master_memory 1024 -master_vcores 2 -num_containers 1 "
            "-container_resources memory-mb={},vcores={} "
            "-shell_script /tmp/{}".format(
                resources.kwargs.get("yarn_path", ""),
                resources.kwargs.get("yarn_path", ""),
                resources.queue_name,
                job.job_hash,
                resources.kwargs.get("img_name", ""),
                resources.kwargs.get("mem_limit", 1) * 1024,
                resources.cpu_per_node,
                script_file_name,
            )
        )

        cmd = "{{ nohup {} 1>{} 2>{} & }} && echo $!".format(
            submit_command,
            output_name,
            output_name,
        )
        ret, stdout, stderr = run_cmd_with_all_output(cmd)

        if ret != 0:
            err_str = stderr.decode("utf-8")
            raise RuntimeError(
                "Command squeue fails to execute, error message:%s\nreturn code %d\n"
                % (err_str, ret)
            )
        job_id = int(stdout.decode("utf-8").strip())

        self.context.write_file(job_id_name, str(job_id))
        return job_id

    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted

        ret, stdout, stderr = run_cmd_with_all_output(
            f"if ps -p {job_id} > /dev/null; then echo 1; fi"
        )
        if ret != 0:
            err_str = stderr.decode("utf-8")
            raise RuntimeError(
                "Command fails to execute, error message:%s\nreturn code %d\n"
                % (err_str, ret)
            )

        if_job_exists = bool(stdout.decode("utf-8").strip())
        if self.check_finish_tag(job=job):
            dlog.info(f"job: {job.job_hash} {job.job_id} finished")
            return JobStatus.finished

        if if_job_exists:
            return JobStatus.running
        else:
            return JobStatus.terminated

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)



import shlex
from typing import List

from dargs import Argument

from dpdispatcher import dlog
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.machine import Machine
from dpdispatcher.utils import RetrySignal, retry

lsf_script_header_template = """\
#!/bin/bash -l
#BSUB -e %J.err
#BSUB -o %J.out
{lsf_nodes_line}
{lsf_ptile_line}
{lsf_partition_line}
{lsf_number_gpu_line}"""


class LSF(Machine):
    """LSF batch."""

    def gen_script(self, job):
        lsf_script = super().gen_script(job)
        return lsf_script

    def gen_script_header(self, job):
        resources = job.resources
        script_header_dict = {
            "lsf_nodes_line": "#BSUB -n {number_cores}".format(
                number_cores=resources.number_node * resources.cpu_per_node
            ),
            "lsf_ptile_line": f"#BSUB -R 'span[ptile={resources.cpu_per_node}]'",
            "lsf_partition_line": f"#BSUB -q {resources.queue_name}",
        }
        gpu_usage_flag = resources.kwargs.get("gpu_usage", False)
        gpu_new_syntax_flag = resources.kwargs.get("gpu_new_syntax", False)
        gpu_exclusive_flag = resources.kwargs.get("gpu_exclusive", True)
        custom_gpu_line = resources.kwargs.get("custom_gpu_line", None)
        if not custom_gpu_line:
            if gpu_usage_flag is True:
                if gpu_new_syntax_flag is True:
                    if gpu_exclusive_flag is True:
                        script_header_dict["lsf_number_gpu_line"] = (
                            "#BSUB -gpu 'num={gpu_per_node}:mode=shared:"
                            "j_exclusive=yes'".format(
                                gpu_per_node=resources.gpu_per_node
                            )
                        )
                    else:
                        script_header_dict["lsf_number_gpu_line"] = (
                            "#BSUB -gpu 'num={gpu_per_node}:mode=shared:"
                            "j_exclusive=no'".format(
                                gpu_per_node=resources.gpu_per_node
                            )
                        )
                else:
                    script_header_dict["lsf_number_gpu_line"] = (
                        '#BSUB -R "select[ngpus >0] rusage['
                        'ngpus_excl_p={gpu_per_node}]"'.format(
                            gpu_per_node=resources.gpu_per_node
                        )
                    )
            else:
                script_header_dict["lsf_number_gpu_line"] = ""
        else:
            script_header_dict["lsf_number_gpu_line"] = custom_gpu_line
        lsf_script_header = lsf_script_header_template.format(**script_header_dict)

        return lsf_script_header

    @retry()
    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        self.context.write_file(fname=script_file_name, write_str=script_str)

        try:
            stdin, stdout, stderr = self.context.block_checkcall(
                "cd {} && {} {}".format(
                    shlex.quote(self.context.remote_root),
                    "bsub < ",
                    shlex.quote(script_file_name),
                )
            )
        except RuntimeError as err:
            raise RetrySignal(err) from err

        subret = stdout.readlines()
        job_id = subret[0].split()[1][1:-1]
        self.context.write_file(job_id_name, job_id)
        return job_id

    # TODO: derive abstract methods
    def default_resources(self, resources):
        pass

    def sub_script_cmd(self, res):
        pass

    def sub_script_head(self, res):
        pass

    @retry()
    def check_status(self, job):
        try:
            job_id = job.job_id
        except AttributeError:
            return JobStatus.terminated
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call("bjobs " + job_id)
        err_str = stderr.read().decode("utf-8")
        if ("Job <%s> is not found" % job_id) in err_str:
            if self.check_finish_tag(job):
                return JobStatus.finished
            else:
                return JobStatus.terminated
        elif ret != 0:
            # just retry when any unknown error raised.
            raise RetrySignal(
                "Get error code %d in checking status through ssh with job: %s . message: %s"
                % (ret, job.job_hash, err_str)
            )
        status_out = stdout.read().decode("utf-8").split("\n")
        if len(status_out) < 2:
            return JobStatus.unknown
        else:
            status_line = status_out[1]
            status_word = status_line.split()[2]

        # ref: https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.2/lsf_command_ref/bjobs.1.html
        if status_word in ["PEND", "WAIT", "PSUSP"]:
            return JobStatus.waiting
        elif status_word in ["RUN", "USUSP"]:
            return JobStatus.running
        elif status_word in ["DONE", "EXIT"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_custom_gpu_line = "Custom GPU configuration, starting with #BSUB"
        doc_gpu_usage = "Choosing if GPU is used in the calculation step. "
        doc_gpu_new_syntax = (
            "For LFS >= 10.1.0.3, new option -gpu for #BSUB could be used. "
            "If False, and old syntax would be used."
        )
        doc_gpu_exclusive = (
            "Only take effect when new syntax enabled. "
            "Control whether submit tasks in exclusive way for GPU."
        )
        return [
            Argument(
                "kwargs",
                dict,
                [
                    Argument(
                        "gpu_usage",
                        bool,
                        optional=True,
                        default=False,
                        doc=doc_gpu_usage,
                    ),
                    Argument(
                        "gpu_new_syntax",
                        bool,
                        optional=True,
                        default=False,
                        doc=doc_gpu_new_syntax,
                    ),
                    Argument(
                        "gpu_exclusive",
                        bool,
                        optional=True,
                        default=True,
                        doc=doc_gpu_exclusive,
                    ),
                    Argument(
                        "custom_gpu_line",
                        str,
                        optional=True,
                        default=None,
                        doc=doc_custom_gpu_line,
                    ),
                ],
                optional=False,
                doc="Extra arguments.",
            )
        ]

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        ret, stdin, stdout, stderr = self.context.block_call("bkill " + str(job_id))



"""Test `Submission.generate_jobs` with different group size."""

import json
import os
import sys
from pathlib import Path
from unittest import TestCase

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import (
    Machine,
    Resources,
    Submission,
    Task,
    setUpModule,  # noqa: F401
)

# 99 tasks in total
# group_size - expected_ntasks
group_ntasks_pairs = [
    (1, 99),
    (3, 33),
    (10, 10),
    (100, 1),
    (0, 1),
]

cwd = Path(__file__).parent
with open(cwd / "jsons" / "machine.json") as f:
    j_machine = json.load(f)["machine"]
with open(cwd / "jsons" / "resources.json") as f:
    j_resources = json.load(f)
with open(cwd / "jsons" / "task.json") as f:
    j_task = json.load(f)


class TestGroupSize(TestCase):
    def test_works_as_expected(self):
        for group_size, ntasks in group_ntasks_pairs:
            with self.subTest(group_size):
                machine = Machine.load_from_dict(j_machine)
                j_resources["group_size"] = group_size
                resources = Resources.load_from_dict(j_resources)
                tasks = [Task.load_from_dict(j_task) for _ in range(99)]
                submission = Submission(".", machine, resources, task_list=tasks)
                submission.generate_jobs()
                self.assertEqual(len(submission.belonging_jobs), ntasks)



import math
import pathlib
import shlex
from typing import List

from dargs import Argument

from dpdispatcher import dlog
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.machine import Machine, script_command_template
from dpdispatcher.utils import RetrySignal, retry

# from dpdispatcher.submission import Resources

slurm_script_header_template = """\
#!/bin/bash -l
#SBATCH --parsable
{slurm_nodes_line}
{slurm_ntasks_per_node_line}
{slurm_number_gpu_line}
{slurm_partition_line}"""


class Slurm(Machine):
    def gen_script(self, job):
        slurm_script = super().gen_script(job)
        return slurm_script

    def gen_script_header(self, job):
        resources = job.resources
        script_header_dict = {}
        script_header_dict["slurm_nodes_line"] = "#SBATCH --nodes {number_node}".format(
            number_node=resources.number_node
        )
        script_header_dict[
            "slurm_ntasks_per_node_line"
        ] = f"#SBATCH --ntasks-per-node {resources.cpu_per_node}"
        custom_gpu_line = resources.kwargs.get("custom_gpu_line", None)
        if not custom_gpu_line:
            script_header_dict[
                "slurm_number_gpu_line"
            ] = f"#SBATCH --gres=gpu:{resources.gpu_per_node}"
        else:
            script_header_dict["slurm_number_gpu_line"] = custom_gpu_line
        if resources.queue_name != "":
            script_header_dict[
                "slurm_partition_line"
            ] = f"#SBATCH --partition {resources.queue_name}"
        else:
            script_header_dict["slurm_partition_line"] = ""
        slurm_script_header = slurm_script_header_template.format(**script_header_dict)
        return slurm_script_header

    @retry()
    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        ret, stdin, stdout, stderr = self.context.block_call(
            "cd {} && {} {}".format(
                shlex.quote(self.context.remote_root),
                "sbatch",
                shlex.quote(script_file_name),
            )
        )
        if ret != 0:
            err_str = stderr.read().decode("utf-8")
            if (
                "Socket timed out on send/recv operation" in err_str
                or "Unable to contact slurm controller" in err_str
            ):
                # server network error, retry 3 times
                raise RetrySignal(
                    "Get error code %d in submitting through ssh with job: %s . message: %s"
                    % (ret, job.job_hash, err_str)
                )
            elif (
                "Job violates accounting/QOS policy" in err_str
                # the number of jobs exceeds DEFAULT_MAX_JOB_COUNT (by default 10000)
                or "Slurm temporarily unable to accept job, sleeping and retrying"
                in err_str
            ):
                # job number exceeds, skip the submitting
                return ""
            raise RuntimeError(
                "status command squeue fails to execute\nerror message:%s\nreturn code %d\n"
                % (err_str, ret)
            )
        subret = stdout.readlines()
        # --parsable
        # Outputs only the job id number and the cluster name if present.
        # The values are separated by a semicolon. Errors will still be displayed.
        job_id = subret[0].split(";")[0].strip()
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    @retry()
    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call(
            'squeue -o "%.18i %.2t" -j ' + job_id
        )
        if ret != 0:
            err_str = stderr.read().decode("utf-8")
            if "Invalid job id specified" in err_str:
                if self.check_finish_tag(job):
                    dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            elif (
                "Socket timed out on send/recv operation" in err_str
                or "Unable to contact slurm controller" in err_str
                or "Invalid user for SlurmUser" in err_str
            ):
                # retry 3 times
                raise RetrySignal(
                    "Get error code %d in checking status through ssh with job: %s . message: %s"
                    % (ret, job.job_hash, err_str)
                )
            raise RuntimeError(
                "status command squeue fails to execute."
                "job_id:%s \n error message:%s\n return code %d\n"
                % (job_id, err_str, ret)
            )
        status_line = stdout.read().decode("utf-8").split("\n")[-2]
        status_word = status_line.split()[-1]
        if not (len(status_line.split()) == 2 and status_word.isupper()):
            raise RuntimeError(
                "Error in getting job status, "
                + f"status_line = {status_line}, "
                + f"parsed status_word = {status_word}"
            )
        if status_word in ["PD", "CF", "S"]:
            return JobStatus.waiting
        elif status_word in ["R"]:
            return JobStatus.running
        elif status_word in ["CG"]:
            return JobStatus.completing
        elif status_word in [
            "C",
            "E",
            "K",
            "BF",
            "CA",
            "CD",
            "F",
            "NF",
            "PR",
            "SE",
            "ST",
            "TO",
        ]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_custom_gpu_line = "Custom GPU configuration, starting with #SBATCH"
        return [
            Argument(
                "kwargs",
                dict,
                [
                    Argument(
                        "custom_gpu_line",
                        str,
                        optional=True,
                        default=None,
                        doc=doc_custom_gpu_line,
                    )
                ],
                optional=True,
                doc="Extra arguments.",
            )
        ]

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        # -Q Do not report an error if the specified job is already completed.
        ret, stdin, stdout, stderr = self.context.block_call(
            "scancel -Q " + str(job_id)
        )
        # we do not need to stop here if scancel failed; just continue


class SlurmJobArray(Slurm):
    """Slurm with job array enabled for multiple tasks in a job."""

    def gen_script_header(self, job):
        slurm_job_size = job.resources.kwargs.get("slurm_job_size", 1)
        if job.fail_count > 0:
            # resubmit jobs, check if some of tasks have been finished
            job_array = set()
            for ii, task in enumerate(job.job_task_list):
                task_tag_finished = (
                    pathlib.PurePath(task.task_work_path)
                    / (task.task_hash + "_task_tag_finished")
                ).as_posix()
                if not self.context.check_file_exists(task_tag_finished):
                    job_array.add(ii // slurm_job_size)
            return super().gen_script_header(job) + "\n#SBATCH --array=%s" % (
                ",".join(map(str, job_array))
            )
        return super().gen_script_header(job) + "\n#SBATCH --array=0-%d" % (
            math.ceil(len(job.job_task_list) / slurm_job_size) - 1
        )

    def gen_script_command(self, job):
        resources = job.resources
        slurm_job_size = resources.kwargs.get("slurm_job_size", 1)
        # SLURM_ARRAY_TASK_ID: 0 ~ n_jobs-1
        script_command = "case $SLURM_ARRAY_TASK_ID in\n"
        for ii, task in enumerate(job.job_task_list):
            command_env = ""
            command_env += self.gen_command_env_cuda_devices(resources=resources)

            task_tag_finished = task.task_hash + "_task_tag_finished"

            log_err_part = ""
            if task.outlog is not None:
                log_err_part += f"1>>{shlex.quote(task.outlog)} "
            if task.errlog is not None:
                log_err_part += f"2>>{shlex.quote(task.errlog)} "

            flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"
            single_script_command = script_command_template.format(
                flag_if_job_task_fail=flag_if_job_task_fail,
                command_env=command_env,
                task_work_path=shlex.quote(
                    pathlib.PurePath(task.task_work_path).as_posix()
                ),
                command=task.command,
                task_tag_finished=task_tag_finished,
                log_err_part=log_err_part,
            )
            if ii % slurm_job_size == 0:
                script_command += f"{ii // slurm_job_size})\n"
            script_command += single_script_command
            script_command += self.gen_script_wait(resources=resources)
            script_command += "\n"
            if (
                ii % slurm_job_size == slurm_job_size - 1
                or ii == len(job.job_task_list) - 1
            ):
                script_command += ";;\n"
        script_command += "*)\nexit 1\n;;\nesac\n"
        return script_command

    def gen_script_end(self, job):
        # We cannot have a end script for job array
        # we may check task tag instead
        return ""

    @retry()
    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call(
            'squeue -h -o "%.18i %.2t" -j ' + job_id
        )
        if ret != 0:
            err_str = stderr.read().decode("utf-8")
            if "Invalid job id specified" in err_str:
                if self.check_finish_tag(job):
                    dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            elif (
                "Socket timed out on send/recv operation" in err_str
                or "Unable to contact slurm controller" in err_str
            ):
                # retry 3 times
                raise RetrySignal(
                    "Get error code %d in checking status through ssh with job: %s . message: %s"
                    % (ret, job.job_hash, err_str)
                )
            raise RuntimeError(
                "status command squeue fails to execute."
                "job_id:%s \n error message:%s\n return code %d\n"
                % (job_id, err_str, ret)
            )
        status_lines = stdout.read().decode("utf-8").split("\n")[:-1]
        status = []
        for status_line in status_lines:
            status_word = status_line.split()[-1]
            if not (len(status_line.split()) == 2 and status_word.isupper()):
                raise RuntimeError(
                    "Error in getting job status, "
                    + f"status_line = {status_line}, "
                    + f"parsed status_word = {status_word}"
                )
            if status_word in ["PD", "CF", "S"]:
                status.append(JobStatus.waiting)
            elif status_word in ["R"]:
                status.append(JobStatus.running)
            elif status_word in ["CG"]:
                status.append(JobStatus.completing)
            elif status_word in [
                "C",
                "E",
                "K",
                "BF",
                "CA",
                "CD",
                "F",
                "NF",
                "PR",
                "SE",
                "ST",
                "TO",
            ]:
                status.append(JobStatus.finished)
            else:
                status.append(JobStatus.unknown)
        # running if any job is running
        if JobStatus.running in status:
            return JobStatus.running
        elif JobStatus.waiting in status:
            return JobStatus.waiting
        elif JobStatus.completing in status:
            return JobStatus.completing
        elif JobStatus.unknown in status:
            return JobStatus.unknown
        else:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated

    def check_finish_tag(self, job):
        results = []
        for task in job.job_task_list:
            task.get_task_state(self.context)
            results.append(task.task_state == JobStatus.finished)
        return all(results)

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        doc_slurm_job_size = "Number of tasks in a Slurm job"
        arg = super().resources_subfields()[0]
        arg.extend_subfields(
            [
                Argument(
                    "slurm_job_size",
                    int,
                    optional=True,
                    default=1,
                    doc=doc_slurm_job_size,
                ),
            ]
        )
        return [arg]



import getpass
import os
import shutil
import sys
import time
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "shell"
from .context import JobStatus, Shell, SSHContext, SSHSession, setUpModule  # noqa: F401


class TestShell(unittest.TestCase):
    def setUp(self):
        os.makedirs("loc", exist_ok=True)
        os.makedirs("rmt", exist_ok=True)
        os.makedirs("loc/task0", exist_ok=True)
        os.makedirs("loc/task1", exist_ok=True)
        for ii in ["loc/task0", "loc/task1"]:
            with open(os.path.join(ii, "test0"), "w") as fp:
                fp.write(str(uuid.uuid4()))
        port = 22
        try:
            ssh_session = SSHSession(
                {
                    "hostname": "localhost",
                    "port": port,
                    "username": getpass.getuser(),
                    "work_path": os.path.join(os.getcwd(), "rmt"),
                }
            )
        except Exception:
            ssh_session = SSHSession(
                {
                    "hostname": "localhost",
                    "port": 5566,
                    "username": getpass.getuser(),
                    "work_path": os.path.join(os.getcwd(), "rmt"),
                }
            )
        self.ctx = SSHContext("loc", ssh_session)
        self.shell = Shell(self.ctx)

    def tearDown(self):
        shutil.rmtree("loc")
        shutil.rmtree("rmt")
        if os.path.exists("dpgen.log"):
            os.remove("dpgen.log")

    def test_gen_sub_script(self):
        job_dirs = ["task0", "task1"]
        self.shell.context.upload(job_dirs, ["test0"])
        ret = self.shell.sub_script(job_dirs, ["touch test1", "touch test2"])
        self.shell.context.write_file("run.sub", ret)
        # with open('run.sub', 'w') as fp:
        #     fp.write(ret)

    def test_sub_success(self):
        job_dirs = ["task0", "task1"]
        self.shell.context.upload(job_dirs, ["test0"])
        self.shell.submit(job_dirs, ["touch test1", "touch test2"])
        while True:
            ret = self.shell.check_status()
            if ret == JobStatus.finished:
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.shell.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.shell.context.remote_root,
                    "%s_tag_finished" % self.shell.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task0/test2")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.shell.context.remote_root, "task1/test2")
            )
        )

    # def test_sub_scancel(self) :
    #     job_dirs = ['task0', 'task1']
    #     self.shell.context.upload(job_dirs, ['test0'])
    #     # sub
    #     self.shell.submit(job_dirs, ['touch test1', 'sleep 10'])
    #     while True:
    #         ret = self.shell.check_status()
    #         if ret == JobStatus.finished  :
    #             raise RuntimeError('should not finished')
    #         if ret == JobStatus.running :
    #             # wait for file writing
    #             time.sleep(2)
    #             # kill job
    #             ##################################################
    #             # problematic killing remotly
    #             ##################################################
    #             self.shell.context.kill(self.shell.proc)
    #             break
    #         time.sleep(1)
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_0_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_0_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_1_finished')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'tag_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test1')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test2')))
    #     self.assertFalse(os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test2')))
    #     # sub restart
    #     self.shell.submit(job_dirs, ['rm test1', 'touch test2'], restart = True)
    #     while True:
    #         ret = self.shell.check_status()
    #         if ret == JobStatus.finished  :
    #             break
    #         time.sleep(1)
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_0_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_0_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/tag_1_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'tag_finished')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test1')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task0/test2')))
    #     self.assertTrue (os.path.isfile(os.path.join('rmt', self.shell.context.remote_root, 'task1/test2')))



import shlex

from dpdispatcher import dlog
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.machine import Machine

pbs_script_header_template = """
#!/bin/bash -l
{select_node_line}
#PBS -j oe
{queue_name_line}
"""


class PBS(Machine):
    def gen_script(self, job):
        pbs_script = super().gen_script(job)
        return pbs_script

    def gen_script_header(self, job):
        resources = job.resources
        pbs_script_header_dict = {}
        pbs_script_header_dict[
            "select_node_line"
        ] = f"#PBS -l select={resources.number_node}:ncpus={resources.cpu_per_node}"
        if resources.gpu_per_node != 0:
            pbs_script_header_dict[
                "select_node_line"
            ] += f":ngpus={resources.gpu_per_node}"
        pbs_script_header_dict["queue_name_line"] = f"#PBS -q {resources.queue_name}"
        pbs_script_header = pbs_script_header_template.format(**pbs_script_header_dict)
        return pbs_script_header

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + "_job_id"
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        # script_file_dir = os.path.join(self.context.submission.work_base)
        script_file_dir = self.context.remote_root
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        stdin, stdout, stderr = self.context.block_checkcall(
            "cd {} && {} {}".format(
                shlex.quote(script_file_dir), "qsub", shlex.quote(script_file_name)
            )
        )
        subret = stdout.readlines()
        job_id = subret[0].split()[0]
        self.context.write_file(job_id_name, job_id)
        return job_id

    def default_resources(self, resources):
        pass

    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call("qstat -x " + job_id)
        err_str = stderr.read().decode("utf-8")
        if ret != 0:
            if "qstat: Unknown Job Id" in err_str or "Job has finished" in err_str:
                if self.check_finish_tag(job=job):
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            else:
                raise RuntimeError(
                    "status command qstat fails to execute. erro info: %s return code %d"
                    % (err_str, ret)
                )
        status_line = stdout.read().decode("utf-8").split("\n")[-2]
        status_word = status_line.split()[-2]
        # dlog.info (status_word)
        if status_word in ["Q", "H"]:
            return JobStatus.waiting
        elif status_word in ["R"]:
            return JobStatus.running
        elif status_word in ["C", "E", "K", "F"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        return self.context.check_file_exists(job_tag_finished)

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        ret, stdin, stdout, stderr = self.context.block_call("qdel " + str(job_id))


class Torque(PBS):
    def check_status(self, job):
        job_id = job.job_id
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr = self.context.block_call("qstat -l " + job_id)
        err_str = stderr.read().decode("utf-8")
        if ret != 0:
            if "qstat: Unknown Job Id" in err_str or "Job has finished" in err_str:
                if self.check_finish_tag(job=job):
                    return JobStatus.finished
                else:
                    return JobStatus.terminated
            else:
                raise RuntimeError(
                    "status command qstat fails to execute. erro info: %s return code %d"
                    % (err_str, ret)
                )
        status_line = stdout.read().decode("utf-8").split("\n")[-2]
        status_word = status_line.split()[-2]
        # dlog.info (status_word)
        if status_word in ["Q", "H"]:
            return JobStatus.waiting
        elif status_word in ["R"]:
            return JobStatus.running
        elif status_word in ["C", "E", "K", "F"]:
            if self.check_finish_tag(job):
                dlog.info(f"job: {job.job_hash} {job.job_id} finished")
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def gen_script_header(self, job):
        # ref: https://support.adaptivecomputing.com/wp-content/uploads/2021/02/torque/torque.htm#topics/torque/2-jobs/requestingRes.htm
        resources = job.resources
        pbs_script_header_dict = {}
        pbs_script_header_dict[
            "select_node_line"
        ] = f"#PBS -l nodes={resources.number_node}:ppn={resources.cpu_per_node}"
        if resources.gpu_per_node != 0:
            pbs_script_header_dict["select_node_line"] += ":gpus={gpu_per_node}".format(
                gpu_per_node=resources.gpu_per_node
            )
        pbs_script_header_dict["queue_name_line"] = f"#PBS -q {resources.queue_name}"
        pbs_script_header = pbs_script_header_template.format(**pbs_script_header_dict)
        return pbs_script_header



import asyncio
import os
import random
import shutil
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
import unittest

from .context import (
    Machine,
    Resources,
    Submission,
    Task,
    setUpModule,  # noqa: F401
)


class RunSubmission:
    def setUp(self):
        self.machine_dict = {
            "batch_type": "Shell",
            "context_type": "LocalContext",
            "local_root": "test_run_submission/",
            # /data is mounted in the docker container
            "remote_root": os.path.join(
                os.environ.get("CI_SHARED_SPACE", "/"), "tmp_run_submission"
            ),
            "remote_profile": {},
        }
        self.resources_dict = {
            "number_node": 1,
            "cpu_per_node": 1,
            "gpu_per_node": 0,
            "queue_name": "?",
            "group_size": 2,
        }
        os.makedirs(
            os.path.join(self.machine_dict["local_root"], "test_dir"), exist_ok=True
        )
        os.makedirs(
            os.path.join(self.machine_dict["local_root"], "test_dir", "test space"),
            exist_ok=True,
        )
        with open(
            os.path.join(
                self.machine_dict["local_root"],
                "test_dir",
                "test space",
                "inp space.txt",
            ),
            "w",
        ) as f:
            f.write("inp space")

    def test_run_submission(self):
        machine = Machine.load_from_dict(self.machine_dict)
        resources = Resources.load_from_dict(self.resources_dict)

        task_list = []
        for ii in range(4):
            task = Task(
                command=f"echo dpdispatcher_unittest_{ii}",
                task_work_path="./",
                forward_files=[],
                backward_files=[f"out{ii}.txt"],
                outlog=f"out{ii}.txt",
            )
            task_list.append(task)

        # test space in file name
        task_list.append(
            Task(
                command="echo dpdispatcher_unittest_space",
                task_work_path="test space/",
                forward_files=["inp space.txt"],
                backward_files=["out space.txt"],
                outlog="out space.txt",
            )
        )
        submission = Submission(
            work_base="test_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=[],
            backward_common_files=[],
            task_list=task_list,
        )
        submission.run_submission(check_interval=2)

        for ii in range(4):
            self.assertTrue(
                os.path.isfile(
                    os.path.join(
                        self.machine_dict["local_root"], "test_dir/", f"out{ii}.txt"
                    )
                )
            )

    def test_async_run_submission(self):
        machine = Machine.load_from_dict(self.machine_dict)
        resources = Resources.load_from_dict(self.resources_dict)
        ntask = 4

        async def run_jobs(ntask):
            background_tasks = set()
            for ii in range(ntask):
                sleep_time = random.random() * 5 + 2
                task = Task(
                    command=f"echo dpdispatcher_unittest_{ii} && sleep {sleep_time}",
                    task_work_path="./",
                    forward_files=[],
                    backward_files=[f"out{ii}.txt"],
                    outlog=f"out{ii}.txt",
                )
                submission = Submission(
                    work_base="test_dir/",
                    machine=machine,
                    resources=resources,
                    forward_common_files=[],
                    backward_common_files=[],
                    task_list=[task],
                )
                background_task = asyncio.create_task(
                    submission.async_run_submission(check_interval=2, clean=False)
                )
                background_tasks.add(background_task)
                # background_task.add_done_callback(background_tasks.discard)
            res = await asyncio.gather(*background_tasks)
            return res

        res = asyncio.run(run_jobs(ntask=ntask))
        print(res)

        for ii in range(4):
            self.assertTrue(
                os.path.isfile(
                    os.path.join(
                        self.machine_dict["local_root"], "test_dir/", f"out{ii}.txt"
                    )
                )
            )

    def tearDown(self):
        shutil.rmtree(os.path.join(self.machine_dict["local_root"]))


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "slurm",
    "outside the slurm testing environment",
)
class TestSlurmRun(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict["batch_type"] = "Slurm"
        self.resources_dict["queue_name"] = "normal"

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "slurm",
    "outside the slurm testing environment",
)
class TestSlurmJobArrayRun(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict["batch_type"] = "SlurmJobArray"
        self.resources_dict["queue_name"] = "normal"

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "slurm",
    "outside the slurm testing environment",
)
class TestSlurmJobArrayRun2(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict["batch_type"] = "SlurmJobArray"
        self.resources_dict["queue_name"] = "normal"
        self.resources_dict["kwargs"] = {"slurm_job_size": 2}

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()


@unittest.skipIf(
    os.environ.get("DPDISPATCHER_TEST") != "pbs", "outside the pbs testing environment"
)
class TestPBSRun(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict["batch_type"] = "PBS"
        self.resources_dict["queue_name"] = "workq"

    @unittest.skip("Manaually skip")  # comment this line to open unittest
    def test_async_run_submission(self):
        return super().test_async_run_submission()


@unittest.skipIf(sys.platform == "win32", "Shell is not supported on Windows")
class TestLazyLocalContext(RunSubmission, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.machine_dict["context_type"] = "LazyLocalContext"



import logging
import os
import sys
import warnings

ROOT_PATH = tuple(__path__)[0]
dlog = logging.getLogger(__name__)
dlog.propagate = False
dlog.setLevel(logging.INFO)
try:
    dlogf = logging.FileHandler(
        os.getcwd() + os.sep + "dpdispatcher" + ".log", delay=True
    )
except PermissionError:
    warnings.warn(
        "dpdispatcher.log meet permission error. redirect the log to ~/dpdispatcher.log"
    )
    dlogf = logging.FileHandler(
        os.path.join(os.path.expanduser("~"), "dpdispatcher.log")
    )

# dlogf = logging.FileHandler('./'+os.sep+SHORT_CMD+'.log')
# dlogf = logging.FileHandler(os.path.join(os.environ['HOME'], SHORT_CMD+'.log'))
# dlogf = logging.FileHandler(os.path.join(os.path.expanduser('~'), SHORT_CMD+'.log'))
# dlogf = logging.FileHandler(os.path.join("/tmp/", SHORT_CMD+'.log'))
dlogf_formatter = logging.Formatter("%(asctime)s - %(levelname)s : %(message)s")
# dlogf_formatter=logging.Formatter('%(asctime)s - %(name)s - [%(filename)s:%(funcName)s - %(lineno)d ] - %(levelname)s \n %(message)s')
dlogf.setFormatter(dlogf_formatter)
dlog.addHandler(dlogf)

dlog_stdout = logging.StreamHandler(sys.stdout)
dlog_stdout.setFormatter(dlogf_formatter)
dlog.addHandler(dlog_stdout)

__author__ = "DeepModeling Team"
__copyright__ = "Copyright 2019"
__status__ = "Development"
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unkown"

from .distributed_shell import DistributedShell
from .dp_cloud_server import DpCloudServer, Lebesgue
from .dp_cloud_server_context import DpCloudServerContext, LebesgueContext
from .fugaku import Fugaku
from .hdfs_context import HDFSContext
from .lazy_local_context import LazyLocalContext
from .local_context import LocalContext
from .lsf import LSF
from .machine import Machine
from .openapi import OpenAPI
from .openapi_context import OpenAPIContext
from .pbs import PBS, Torque
from .shell import Shell
from .slurm import Slurm
from .ssh_context import SSHContext
from .submission import Job, Resources, Submission, Task


def info():
    """Show basic information about dpdispatcher, its location and version."""
    print("DeepModeling\n------------")
    print("Version: " + __version__)
    print("Path:    " + ROOT_PATH)
    print("")
    print("Dependency")
    print("------------")
    for modui in ["psutil", "paramiko", "dargs", "oss2"]:
        try:
            mm = __import__(modui)
            print("%10s %10s   %s" % (modui, mm.__version__, mm.__path__[0]))
        except ImportError:
            print("%10s %10s Not Found" % (modui, ""))
    print()


__all__ = [
    "__version__",
    "DistributedShell",
    "DpCloudServer",
    "OpenAPI",
    "OpenAPIContext",
    "DpCloudServerContext",
    "HDFSContext",
    "LazyLocalContext",
    "LocalContext",
    "LSF",
    "Machine",
    "PBS",
    "Shell",
    "Slurm",
    "Fugaku",
    "SSHContext",
    "Submission",
    "Task",
    "Torque",
    "info",
    "Lebesgue",
    "LebesgueContext",
    "Job",
    "Resources",
]



import json
import os
import sys

from dpdispatcher.machine import Machine
from dpdispatcher.submission import Resources, Submission, Task

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

# task_need_resources has no effect
with open("jsons/machine_lsf.json") as f:
    mdata = json.load(f)

machine = Machine.load_from_dict(mdata["machine"])
resources = Resources.load_from_dict(mdata["resources"])

submission = Submission(
    work_base="0_md/",
    machine=machine,
    resources=resources,
    forward_common_files=["graph.pb"],
    backward_common_files=[],
)

task1 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-1/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task2 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-2/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task3 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-3/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
task4 = Task(
    command="lmp -i input.lammps",
    task_work_path="bct-4/",
    forward_files=["conf.lmp", "input.lammps"],
    backward_files=["log.lammps"],
)
submission.register_task_list(
    [
        task1,
        task2,
        task3,
        task4,
    ]
)
submission.run_submission(clean=True)



# %%
# import sys, os
# sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..' )))
# import dpdispatcher
from dpdispatcher.machine import Machine
from dpdispatcher.submission import Resources, Task

# %%
resources_dargs_doc = Resources.arginfo().gen_doc()
with open("../doc/resources-auto.rst", "w") as f:
    # print(resources_dargs_doc)
    f.write(resources_dargs_doc)

machine_dargs_doc = Machine.arginfo().gen_doc()
with open("../doc/machine-auto.rst", "w") as f:
    f.write(machine_dargs_doc)

task_dargs_doc = Task.arginfo().gen_doc()
with open("../doc/task-auto.rst", "w") as f:
    f.write(task_dargs_doc)


# %%



import os
import re
import time
import urllib.parse
from urllib.parse import urljoin

import requests

from dpdispatcher import dlog

from .config import API_HOST, API_LOGGER_STACK_INFO, HTTP_TIME_OUT
from .retcode import RETCODE

try:
    import oss2
    from oss2 import SizedFileAdapter, determine_part_size
    from oss2.models import PartInfo
except ImportError:
    pass
ENABLE_STACK = True if API_LOGGER_STACK_INFO else False


class RequestInfoException(Exception):
    pass


class Client:
    def __init__(
        self, email=None, password=None, debug=False, ticket=None, base_url=API_HOST
    ):
        self.debug = debug
        self.debug = os.getenv("LBG_CLI_DEBUG_PRINT", debug)
        self.config = {}
        self.token = ""
        self.user_id = None
        self.config["email"] = email
        self.config["password"] = password
        self.base_url = base_url
        self.last_log_offset = 0
        self.ticket = ticket

    def post(self, url, data=None, header=None, params=None, retry=5):
        return self._req(
            "POST", url, data=data, header=header, params=params, retry=retry
        )

    def get(self, url, header=None, params=None, retry=5):
        return self._req("GET", url, header=header, params=params, retry=retry)

    def _req(self, method, url, data=None, header=None, params=None, retry=5):
        short_url = url
        url = urllib.parse.urljoin(self.base_url, url)
        if header is None:
            header = {}
        if not self.token:
            self.refresh_token()
        self.ticket = os.environ.get("BOHR_TICKET", "")
        header["Authorization"] = f"jwt {self.token}"
        header["Brm-Ticket"] = self.ticket
        resp_code = None
        err = None
        for i in range(retry):
            resp = None
            try:
                if method == "GET":
                    resp = requests.get(url, params=params, headers=header)
                else:
                    if self.debug:
                        print(data)
                    resp = requests.post(url, json=data, params=params, headers=header)
            except Exception as e:
                dlog.error(f"request({i}) error {e}", i, stack_info=ENABLE_STACK)
                err = e
                time.sleep(1 * i)
                continue

            resp_code = resp.status_code
            if not resp.ok:
                if self.debug:
                    print(f"retry: {i},statusCode: {resp.status_code}")
                try:
                    result = resp.json()
                    err = result.get("error")
                except Exception:
                    pass
                time.sleep(0.1 * i)
                continue
            result = resp.json()
            if result["code"] == "0000" or result["code"] == 0:
                return result.get("data", {})
            else:
                self.token = ""
                self.refresh_token()
                err = result.get("message") or result.get("error")
        raise RequestInfoException(resp_code, short_url, err)

    def _login(self):
        if self.config["email"] is None or self.config["password"] is None:
            raise RequestInfoException(
                "can not find login information, please check your config"
            )
        post_data = {"email": self.config["email"], "password": self.config["password"]}
        resp = self.post("/account/login", post_data)
        self.token = resp["token"]
        # print(self.token)
        self.user_id = resp["user_id"]

    def refresh_token(self, retry=3):
        self.ticket = os.environ.get("BOHR_TICKET", "")
        if self.ticket:
            return
        url = "/account/login"
        post_data = {"email": self.config["email"], "password": self.config["password"]}
        resp_code = None
        err = None
        for i in range(retry):
            resp = requests.post(
                urljoin(API_HOST, url),
                json=post_data,
                timeout=HTTP_TIME_OUT,
            )
            resp_code = resp.status_code
            if not resp.ok:
                if self.debug:
                    print(f"login retry: {i},statusCode: {resp.status_code}")
                try:
                    result = resp.json()
                    err = result.get("error")
                except Exception:
                    pass
                time.sleep(1 * i)
                continue

            result = resp.json()
            if result["code"] == RETCODE.OK or result["code"] == 0:
                self.token = result["data"]["token"]
                return
        raise RequestInfoException(resp_code, url, err)

    def _get_oss_bucket(self, endpoint, bucket_name):
        #  res = get("/tools/sts_token", {})
        res = self.get("/data/get_sts_token", {})
        # print('debug>>>>>>>>>>>>>', res)
        dlog.debug(f"debug: _get_oss_bucket: res:{res}")
        auth = oss2.StsAuth(
            res["AccessKeyId"], res["AccessKeySecret"], res["SecurityToken"]
        )
        return oss2.Bucket(auth, endpoint, bucket_name)

    def download(self, oss_file, save_file, endpoint, bucket_name):
        bucket = self._get_oss_bucket(endpoint, bucket_name)
        dlog.debug(f"debug: download: oss_file:{oss_file}; save_file:{save_file}")
        bucket.get_object_to_file(oss_file, save_file)
        return save_file

    def download_from_url(self, url, save_file):
        ret = None
        for retry_count in range(3):
            try:
                ret = requests.get(
                    url, headers={"Authorization": "jwt " + self.token}, stream=True
                )
            except Exception as e:
                dlog.error(f"request error {e}", stack_info=ENABLE_STACK)
                continue
            if ret.ok:
                break
            else:
                dlog.error(
                    f"request error status_code:{ret.status_code} reason: {ret.reason} body: \n{ret.text}"
                )
                time.sleep(retry_count)
                ret = None
        if ret is not None:
            ret.raise_for_status()
            with open(save_file, "wb") as f:
                for chunk in ret.iter_content(chunk_size=8192):
                    f.write(chunk)
            ret.close()

    def upload(self, oss_task_zip, zip_task_file, endpoint, bucket_name):
        dlog.debug(
            f"debug: upload: oss_task_zip:{oss_task_zip}; zip_task_file:{zip_task_file}"
        )
        bucket = self._get_oss_bucket(endpoint, bucket_name)
        total_size = os.path.getsize(zip_task_file)
        part_size = determine_part_size(total_size, preferred_size=1000 * 1024)
        upload_id = bucket.init_multipart_upload(oss_task_zip).upload_id
        parts = []
        with open(zip_task_file, "rb") as fileobj:
            part_number = 1
            offset = 0
            while offset < total_size:
                num_to_upload = min(part_size, total_size - offset)
                result = bucket.upload_part(
                    oss_task_zip,
                    upload_id,
                    part_number,
                    SizedFileAdapter(fileobj, num_to_upload),
                )
                parts.append(PartInfo(part_number, result.etag))
                offset += num_to_upload
                part_number += 1
        # result = bucket.complete_multipart_upload(oss_task_zip, upload_id, parts)
        result = bucket.complete_multipart_upload(oss_task_zip, upload_id, parts)
        # print('debug:upload_result:', result, dir())
        return result

    def job_create(
        self, job_type, oss_path, input_data, program_id=None, group_id=None
    ):
        post_data = {
            "job_type": job_type,
            "oss_path": oss_path,
        }
        if program_id is not None:
            post_data["project_id"] = program_id
        if group_id is not None:
            post_data["job_group_id"] = group_id
        for k, v in input_data.items():
            post_data[k] = v
        if input_data.get("backward_files"):
            post_data["out_files"] = input_data.get("backward_files")
        if input_data.get("command"):
            post_data["cmd"] = input_data.get("command")
        if input_data.get("machine_type"):
            post_data["scass_type"] = input_data.get("machine_type")
        log = input_data.get(
            "logFiles", input_data.get("log_files", input_data.get("log_file"))
        )
        if log:
            if isinstance(log, str):
                post_data["log_files"] = [log]
        if (
            "checkpoint_files" in post_data
            and post_data["checkpoint_files"] == "sync_files"
        ):
            post_data["checkpoint_files"] = ["*"]
        camel_data = {self._camelize(k): v for k, v in post_data.items()}
        ret = self.post("/brm/v2/job/add", camel_data)
        group_id = ret.get("jobGroupId")
        return ret["jobId"], group_id

    def _camelize(self, str_or_iter):
        # code reference from https://pypi.org/project/pyhumps/
        regex = re.compile(r"(?<=[^\-_\s])[\-_\s]+[^\-_\s]")

        def _is_none(_in):
            return "" if _in is None else _in

        s = str(_is_none(str_or_iter))
        if s.isupper() or s.isnumeric():
            return str_or_iter

        if len(s) != 0 and not s[:2].isupper():
            s = s[0].lower() + s[1:]
        return regex.sub(lambda m: m.group(0)[-1].upper(), s)

    def get_job_detail(self, job_id):
        try:
            ret = self.get(
                f"brm/v1/job/{job_id}",
            )
        except RequestInfoException as e:
            if e.args[0] != 200:
                raise e

            dlog.error(f"get job detail error {e}", stack_info=ENABLE_STACK)
            return None

        return ret

    def get_log(self, job_id):
        url, size = self._get_job_log(job_id)
        if not url:
            return ""
        if self.last_log_offset >= size:
            return ""
        resp = requests.get(url, headers={"Range": f"bytes={self.last_log_offset}-"})
        self.last_log_offset += len(resp.content)
        return resp.content.decode("utf-8")

    def _get_job_log(self, job_id):
        ret = self.get(
            f"/brm/v1/job/{job_id}/log",
            params={
                "pageSize": 1,
            },
        )
        d = ret.get("logFiles")
        if d and len(d) != 0:
            return d[0]["url"], d[0]["size"]
        return None, 0

    def get_tasks_list(self, group_id, per_page=30):
        result = []
        page = 0
        while True:
            ret = self.get(
                "/brm/v1/job/list",
                params={
                    "groupId": group_id,
                    "page": page,
                    "pageSize": per_page,
                },
            )
            if len(ret["items"]) == 0:
                break
            for each in ret["items"]:
                result.append(each)
            page += 1
        return result

    def get_job_result_url(self, job_id):
        try:
            if not job_id:
                return None
            if "job_group_id" in job_id:
                ids = job_id.split(":job_group_id:")
                job_id, _ = int(ids[0]), int(ids[1])
            ret = self.get(f"/brm/v1/job/{job_id}", {})
            if "resultUrl" in ret and len(ret["resultUrl"]) != 0:
                return ret.get("resultUrl")
            else:
                return None
        except ValueError as e:
            dlog.error(e, stack_info=ENABLE_STACK)
            return None



# %%
import os
import sys
import unittest
import uuid

from dpdispatcher.dpcloudserver import api
from dpdispatcher.dpcloudserver.zip_file import zip_files

# import api
# from zip_file import zip_files


# %%
class DPTest(unittest.TestCase):
    test_data = {
        "job_type": "indicate",
        "log_file": "mylog",
        "command": "( echo aa && lmp -i input.lammps && sleep 900 ) > dp.log 2>&1",
        "backward_files": [],
        "job_name": "dpdispatcher_lammps_test",
        "machine": {
            "platform": "ali",
            "resources": {
                "gpu_type": "1 * NVIDIA P100",
                "cpu_num": 4,
                "mem_limit": 28,
                "time_limit": "2:00:00",
                "image_name": "yfb-deepmd-kit-1.2.4-cuda10",
            },
        },
        "job_resources": "http://dpcloudserver.oss-cn-shenzhen.aliyuncs.com/dpcloudserver/indicate/a657ff49722839f1ee54edeb3e9b1beb0ee5cc0e/a657ff49722839f1ee54edeb3e9b1beb0ee5cc0e.zip",
    }

    username = ""
    password = ""

    ENDPOINT = "http://oss-cn-shenzhen.aliyuncs.com"
    BUCKET_NAME = "dpcloudserver"

    @classmethod
    def setUpClass(cls):
        print("execute", sys._getframe().f_code.co_name)

    @classmethod
    def tearDownClass(cls):
        print("execute", sys._getframe().f_code.co_name)

    def setUp(self):
        print("execute", sys._getframe().f_code.co_name)
        api.login(self.username, self.password)

    def test_commit_job(self):
        print("----------", sys._getframe().f_code.co_name)
        file_uuid = uuid.uuid1().hex
        oss_task_zip = os.path.join(
            "{}/{}/{}.zip".format("indicate", file_uuid, file_uuid)
        )
        zip_path = "/home/felix/workplace/22_dpdispatcher/dpdispatcher-yfb/dpdispatcher/dpcloudserver/t.txt"
        zip_task_file = zip_path + ".zip"
        zip_files(zip_path, zip_task_file, [])
        api.upload(oss_task_zip, zip_task_file, self.ENDPOINT, self.BUCKET_NAME)
        job_id = api.job_create(
            self.test_data["job_type"], self.test_data["job_resources"], self.test_data
        )
        tasks = api.get_job_detail(job_id)
        print(tasks)

    def test_get_tasks(self):
        print("----------", sys._getframe().f_code.co_name)
        jobs = api.get_jobs()
        for j in jobs:
            tasks = api.get_job_detail(j["id"])
            print(tasks)

    # def test_download(self):
    #     print('----------', sys._getframe().f_code.co_name)
    #     oss_path = 'dpcloudserver/indicate/abe0febc92ce11eb990800163e094dc5/abe0febc92ce11eb990800163e094dc5.zip'
    #     api.download(oss_path, "out.zip", self.ENDPOINT, self.BUCKET_NAME)


if __name__ == "__main__":
    suite = unittest.TestSuite()

    suite.addTest(DPTest("test_commit_job"))
    # suite.addTest(DPTest("test_get_tasks"))
    # suite.addTest(DPTest("test_download"))

    runner = unittest.TextTestRunner()
    runner.run(suite)



import json
import pathlib
import shlex
from abc import ABCMeta, abstractmethod
from typing import List, Tuple

from dargs import Argument, Variant

from dpdispatcher import dlog
from dpdispatcher.base_context import BaseContext

script_template = """\
{script_header}
{script_custom_flags}
{script_env}
{script_command}
{script_end}
"""

script_env_template = """
REMOTE_ROOT=$(readlink -f {remote_root})
echo 0 > $REMOTE_ROOT/{flag_if_job_task_fail}
test $? -ne 0 && exit 1

{module_unload_part}
{module_load_part}
{source_files_part}
{export_envs_part}
{prepend_script_part}
"""

script_command_template = """
cd $REMOTE_ROOT
cd {task_work_path}
test $? -ne 0 && exit 1
if [ ! -f {task_tag_finished} ] ;then
  {command_env} ( {command} ) {log_err_part}
  if test $? -eq 0; then touch {task_tag_finished}; else echo 1 > $REMOTE_ROOT/{flag_if_job_task_fail};fi
fi &
"""

script_end_template = """
cd $REMOTE_ROOT
test $? -ne 0 && exit 1

wait
FLAG_IF_JOB_TASK_FAIL=$(cat {flag_if_job_task_fail})
if test $FLAG_IF_JOB_TASK_FAIL -eq 0; then touch {job_tag_finished}; else exit 1;fi

{append_script_part}
"""


class Machine(metaclass=ABCMeta):
    """A machine is used to handle the connection with remote machines.

    Parameters
    ----------
    context : SubClass derived from BaseContext
        The context is used to mainatin the connection with remote machine.
    """

    subclasses_dict = {}
    options = set()
    # alias: for subclasses_dict key
    # notes: this attribute can be inherited
    alias: Tuple[str, ...] = tuple()

    def __new__(cls, *args, **kwargs):
        if cls is Machine:
            subcls = cls.subclasses_dict[kwargs["batch_type"]]
            instance = subcls.__new__(subcls, *args, **kwargs)
        else:
            instance = object.__new__(cls)
        return instance

    def __init__(
        self,
        batch_type=None,
        context_type=None,
        local_root=None,
        remote_root=None,
        remote_profile={},
        *,
        context=None,
    ):
        if context is None:
            assert isinstance(self, self.__class__.subclasses_dict[batch_type])
            context = BaseContext(  # type: ignore
                context_type=context_type,
                local_root=local_root,
                remote_root=remote_root,
                remote_profile=remote_profile,
            )
        else:
            pass
        self.bind_context(context=context)

    def bind_context(self, context):
        self.context = context

    # def __init__ (self,
    #             context):
    #     self.context = context
    # self.uuid_names = uuid_names
    # self.upload_tag_name = '%s_job_tag_upload' % self.context.job_uuid
    # self.finish_tag_name = '%s_job_tag_finished' % self.context.job_uuid
    # self.sub_script_name = '%s.sub' % self.context.job_uuid
    # self.job_id_name = '%s_job_id' % self.context.job_uuid

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        alias = [cls.__name__, *cls.alias]
        for aa in alias:
            cls.subclasses_dict[aa] = cls
            cls.subclasses_dict[aa.lower()] = cls
        cls.options.add(cls.__name__)
        # cls.subclasses.append(cls)

    @classmethod
    def load_from_json(cls, json_path):
        with open(json_path) as f:
            machine_dict = json.load(f)
        machine = cls.load_from_dict(machine_dict=machine_dict)
        return machine

    @classmethod
    def load_from_dict(cls, machine_dict):
        batch_type = machine_dict["batch_type"]
        try:
            machine_class = cls.subclasses_dict[batch_type]
        except KeyError as e:
            dlog.error(
                f"KeyError:batch_type; cls.subclasses_dict:{cls.subclasses_dict}; batch_type:{batch_type};"
            )
            raise e
        # check dict
        base = cls.arginfo()
        machine_dict = base.normalize_value(machine_dict, trim_pattern="_*")
        base.check_value(machine_dict, strict=False)

        context = BaseContext.load_from_dict(machine_dict)
        machine = machine_class(context=context)
        return machine

    def serialize(self, if_empty_remote_profile=False):
        machine_dict = {}
        machine_dict["batch_type"] = self.__class__.__name__
        machine_dict["context_type"] = self.context.__class__.__name__
        machine_dict["local_root"] = self.context.init_local_root
        machine_dict["remote_root"] = self.context.init_remote_root
        if not if_empty_remote_profile:
            machine_dict["remote_profile"] = self.context.remote_profile
        else:
            machine_dict["remote_profile"] = {}
        return machine_dict

    def __eq__(self, other):
        return self.serialize() == other.serialize()

    @classmethod
    def deserialize(cls, machine_dict):
        machine = cls.load_from_dict(machine_dict=machine_dict)
        return machine

    @abstractmethod
    def check_status(self, job):
        raise NotImplementedError(
            "abstract method check_status should be implemented by derived class"
        )

    def default_resources(self, res):
        raise NotImplementedError(
            "abstract method default_resources should be implemented by derived class"
        )

    def sub_script_head(self, res):
        raise NotImplementedError(
            "abstract method sub_script_head should be implemented by derived class"
        )

    def sub_script_cmd(self, res):
        raise NotImplementedError(
            "abstract method sub_script_cmd should be implemented by derived class"
        )

    @abstractmethod
    def do_submit(self, job):
        """Submit a single job, assuming that no job is running there."""
        raise NotImplementedError(
            "abstract method do_submit should be implemented by derived class"
        )

    def gen_script(self, job):
        script_header = self.gen_script_header(job)
        script_custom_flags = self.gen_script_custom_flags_lines(job)
        script_env = self.gen_script_env(job)
        script_command = self.gen_script_command(job)
        script_end = self.gen_script_end(job)
        script = script_template.format(
            script_header=script_header,
            script_custom_flags=script_custom_flags,
            script_env=script_env,
            script_command=script_command,
            script_end=script_end,
        )
        return script

    def check_if_recover(self, submission):
        submission_hash = submission.submission_hash
        submission_file_name = f"{submission_hash}.json"
        if_recover = self.context.check_file_exists(submission_file_name)
        return if_recover

    @abstractmethod
    def check_finish_tag(self, **kwargs):
        raise NotImplementedError(
            "abstract method check_finish_tag should be implemented by derived class"
        )

    @abstractmethod
    def gen_script_header(self, job):
        raise NotImplementedError(
            "abstract method gen_script_header should be implemented by derived class"
        )

    def gen_script_custom_flags_lines(self, job):
        custom_flags_lines = ""
        custom_flags = job.resources.custom_flags
        for ii in custom_flags:
            line = ii + "\n"
            custom_flags_lines += line
        return custom_flags_lines

    def gen_script_env(self, job):
        source_files_part = ""

        module_unload_part = ""
        module_purge = job.resources.module_purge
        if module_purge:
            module_unload_part += "module purge\n"
        module_unload_list = job.resources.module_unload_list
        for ii in module_unload_list:
            module_unload_part += f"module unload {ii}\n"

        module_load_part = ""
        module_list = job.resources.module_list
        for ii in module_list:
            module_load_part += f"module load {ii}\n"

        source_list = job.resources.source_list
        for ii in source_list:
            line = "{ source %s; } \n" % ii
            source_files_part += line

        export_envs_part = ""
        envs = job.resources.envs
        for k, v in envs.items():
            if isinstance(v, list):
                for each_value in v:
                    export_envs_part += f"export {k}={each_value}\n"
            else:
                export_envs_part += f"export {k}={v}\n"

        prepend_script = job.resources.prepend_script
        prepend_script_part = "\n".join(prepend_script)

        flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"

        script_env = script_env_template.format(
            flag_if_job_task_fail=flag_if_job_task_fail,
            remote_root=shlex.quote(self.context.remote_root),
            module_unload_part=module_unload_part,
            module_load_part=module_load_part,
            source_files_part=source_files_part,
            export_envs_part=export_envs_part,
            prepend_script_part=prepend_script_part,
        )
        return script_env

    def gen_script_command(self, job):
        script_command = ""
        resources = job.resources
        # in_para_task_num = 0
        for task in job.job_task_list:
            command_env = ""
            command_env += self.gen_command_env_cuda_devices(resources=resources)

            task_tag_finished = task.task_hash + "_task_tag_finished"

            log_err_part = ""
            if task.outlog is not None:
                log_err_part += f"1>>{shlex.quote(task.outlog)} "
            if task.errlog is not None:
                log_err_part += f"2>>{shlex.quote(task.errlog)} "

            flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"
            single_script_command = script_command_template.format(
                flag_if_job_task_fail=flag_if_job_task_fail,
                command_env=command_env,
                task_work_path=shlex.quote(
                    pathlib.PurePath(task.task_work_path).as_posix()
                ),
                command=task.command,
                task_tag_finished=task_tag_finished,
                log_err_part=log_err_part,
            )
            script_command += single_script_command

            script_command += self.gen_script_wait(resources=resources)
        return script_command

    def gen_script_end(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        flag_if_job_task_fail = job.job_hash + "_flag_if_job_task_fail"

        append_script = job.resources.append_script
        append_script_part = "\n".join(append_script)

        script_end = script_end_template.format(
            job_tag_finished=job_tag_finished,
            flag_if_job_task_fail=flag_if_job_task_fail,
            append_script_part=append_script_part,
        )
        return script_end

    def gen_script_wait(self, resources):
        # if not resources.strategy.get('if_cuda_multi_devices', None):
        #     return "wait \n"
        para_deg = resources.para_deg
        resources.task_in_para += 1
        # task_need_gpus = task.task_need_gpus
        if resources.task_in_para >= para_deg:
            # pbs_script_command += pbs_script_wait
            resources.task_in_para = 0
            if resources.strategy["if_cuda_multi_devices"] is True:
                resources.gpu_in_use += 1
                if resources.gpu_in_use % resources.gpu_per_node == 0:
                    return "wait \n"
                else:
                    return ""
            return "wait \n"
        return ""

    def gen_command_env_cuda_devices(self, resources):
        # task_need_resources = task.task_need_resources
        # task_need_gpus = task_need_resources.get('task_need_gpus', 1)
        command_env = ""
        # gpu_number = resources.gpu_per_node
        # resources.gpu_in_use = 0

        if resources.strategy["if_cuda_multi_devices"] is True:
            if resources.gpu_per_node == 0:
                raise RuntimeError("resources.gpu_per_node can not be 0")
            gpu_index = resources.gpu_in_use % resources.gpu_per_node
            command_env += f"export CUDA_VISIBLE_DEVICES={gpu_index};"
            # for ii in list_CUDA_VISIBLE_DEVICES:
            #     command_env+="{ii},".format(ii=ii)
        return command_env

    @classmethod
    def arginfo(cls):
        # TODO: change the possible value of batch and context types after we refactor the code
        doc_batch_type = "The batch job system type. Option: " + ", ".join(cls.options)
        doc_context_type = (
            "The connection used to remote machine. Option: "
            + ", ".join(BaseContext.options)
        )
        doc_local_root = "The dir where the tasks and relating files locate. Typically the project dir."
        doc_remote_root = "The dir where the tasks are executed on the remote machine. Only needed when context is not lazy-local."
        doc_clean_asynchronously = (
            "Clean the remote directory asynchronously after the job finishes."
        )

        machine_args = [
            Argument("batch_type", str, optional=False, doc=doc_batch_type),
            # TODO: add default to local_root and remote_root after refactor the code
            Argument(
                "local_root", [str, type(None)], optional=False, doc=doc_local_root
            ),
            Argument(
                "remote_root", [str, type(None)], optional=True, doc=doc_remote_root
            ),
            Argument(
                "clean_asynchronously",
                bool,
                optional=True,
                default=False,
                doc=doc_clean_asynchronously,
            ),
        ]

        context_variant = Variant(
            "context_type",
            [
                context.machine_arginfo()
                for context in set(BaseContext.subclasses_dict.values())
            ],
            optional=False,
            doc=doc_context_type,
        )

        machine_format = Argument("machine", dict, machine_args, [context_variant])
        return machine_format

    @classmethod
    def resources_arginfo(cls) -> Argument:
        """Generate the resources arginfo.

        Returns
        -------
        Argument
            resources arginfo
        """
        alias = []
        for aa in cls.alias:
            alias.extend(
                (
                    aa,
                    aa.lower(),
                )
            )
        return Argument(
            cls.__name__,
            dict,
            sub_fields=cls.resources_subfields(),
            alias=[cls.__name__.lower(), *alias],
        )

    @classmethod
    def resources_subfields(cls) -> List[Argument]:
        """Generate the resources subfields.

        Returns
        -------
        list[Argument]
            resources subfields
        """
        return [
            Argument(
                "kwargs", dict, optional=True, doc="This field is empty for this batch."
            )
        ]

    def kill(self, job):
        """Kill the job.

        If not implemented, pass and let the user manually kill it.

        Parameters
        ----------
        job : Job
            job
        """
        dlog.warning("Job %s should be manually killed" % job.job_id)
